import csv
import os
from dataclasses import dataclass
from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from .exec_model import ExecModel, ModelObject


@dataclass
class OptimizationResults(ExecModel):
    model: ModelObject

    def __post_init__(self) -> None:
        """
        Create optimization_results/ in the model folder.
        """
        os.makedirs(
            os.path.join(
                self.model.path,
                "optimization_results",
            ),
            exist_ok=True,
        )

    def to_csv(self) -> None:
        """
        Save optimized parameters as CSV file format.

        Examples
        --------
        >>> from biomass import Model, OptimizationResults
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> model = Model(Nakakuki_Cell_2010.__package__).create()
        >>> res = OptimizationResults(model)
        >>> res.to_csv()

        Notes
        -----
        Output:

        * optimization_results/optimized_params.csv
        * optimization_results/optimized_initials.csv
        """
        n_file = self.get_executable()

        if len(self.model.problem.idx_params) + len(self.model.problem.idx_initials) > 0:
            optimized_params = np.empty(
                (
                    len(self.model.problem.idx_params) + len(self.model.problem.idx_initials) + 2,
                    len(n_file) + 1,
                ),
                dtype="<U21",
            )
            for j, nth_paramset in enumerate(sorted(n_file), start=1):
                for i, parameter_index in enumerate(self.model.problem.idx_params):
                    best_generation = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            "generation.npy",
                        )
                    )
                    best_individual = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            f"fit_param{int(best_generation):d}.npy",
                        )
                    )
                    error = np.load(
                        os.path.join(
                            self.model.path,
                            "out",
                            f"{nth_paramset:d}",
                            "best_fitness.npy",
                        )
                    )
                    optimized_params[0, 0] = ""
                    optimized_params[1, 0] = "*Error*"
                    optimized_params[0, j] = str(nth_paramset)
                    optimized_params[1, j] = f"{error:8.3e}"
                    optimized_params[i + 2, 0] = self.model.parameters[parameter_index]
                    optimized_params[i + 2, j] = f"{best_individual[i]:8.3e}"
                if all([isinstance(_s, str) for _s in self.model.species]):
                    for i, species_index in enumerate(self.model.problem.idx_initials):
                        optimized_params[i + len(self.model.problem.idx_params) + 2, 0] = (
                            "init_" + self.model.species[species_index]
                        )
                        optimized_params[
                            i + len(self.model.problem.idx_params) + 2, j
                        ] = f"{best_individual[i+len(self.model.problem.idx_params)]:8.3e}"
            with open(
                os.path.join(
                    self.model.path,
                    "optimization_results",
                    "optimized_params.csv",
                ),
                mode="w",
            ) as f:
                writer = csv.writer(f, lineterminator="\n")
                writer.writerows(optimized_params)

    def savefig(
        self,
        *,
        figsize: Optional[Tuple[float, float]] = None,
        config: Optional[dict] = None,
        boxplot_kws: Optional[dict] = None,
    ) -> None:
        """
        Visualize estimated parameter sets using `seaborn.boxplot`.

        Parameters
        ----------
        figsize : Tuple[float, float], optional
            Width, height in inches.
        config : dict, optional
            A dictionary object for setting `matplotlib.rcParams`.
        boxplot_kws : dict, optional
            Keyword arguments to pass to `seaborn.boxplot`.

        Examples
        --------
        >>> from biomass import Model, OptimizationResults
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> model = Model(Nakakuki_Cell_2010.__package__).create()
        >>> res = OptimizationResults(model)
        >>> res.savefig(figsize=(16,5), boxplot_kws={"orient": "v"})

        Notes
        -----
        Output:

        * optimization_results/estimated_parameter_sets.pdf
        """
        if not os.path.isfile(
            os.path.join(
                self.model.path,
                "optimization_results",
                "optimized_params.csv",
            )
        ):
            self.to_csv()
        if boxplot_kws is None:
            boxplot_kws = {}
        boxplot_kws.setdefault("orient", "v")

        df = pd.read_csv(
            os.path.join(
                self.model.path,
                "optimization_results",
                "optimized_params.csv",
            ),
            index_col=0,
        )
        df.drop("*Error*", inplace=True)

        if config is None:
            config = {}
        if figsize is not None:
            config["figure.figsize"] = figsize
        config.setdefault("savefig.bbox", "tight")
        config.setdefault("savefig.format", "pdf")
        plt.rcParams.update(config)
        ax = sns.boxplot(data=df.T, **boxplot_kws)
        if boxplot_kws["orient"] == "h":
            ax.set_xscale("log")
            ax.set_xlabel("Parameter value")
        elif boxplot_kws["orient"] == "v":
            ax.set_yscale("log")
            ax.set_ylabel("Parameter value")
            ax.set_xticklabels(ax.get_xticklabels(), rotation=90)
        else:
            raise ValueError("boxplot_kws['orient'] must be either 'v' or 'h'.")
        sns.despine()
        plt.savefig(
            os.path.join(
                self.model.path,
                "optimization_results",
                "estimated_parameter_sets",
            ),
        )
        plt.rcdefaults()
        plt.close()

    def dynamic_assessment(self, include_original: bool = False) -> None:
        """
        Compute objective values using estimated parameters.

        Parameters
        ----------
        include_original : bool (default: False)
            If True, an objective value simulated with original parameters
            will also be shown.

        Examples
        --------
        >>> from biomass import Model, OptimizationResults
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> model = Model(Nakakuki_Cell_2010.__package__).create()
        >>> res = OptimizationResults(model)
        >>> res.dynamic_assessment()

        Notes
        -----
        Output:

        * optimization_results/fitness_assessment.csv
        """
        with open(
            os.path.join(
                self.model.path,
                "optimization_results",
                "fitness_assessment.csv",
            ),
            mode="w",
        ) as f:
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(["parameter set", "Objective value"])
            if include_original:
                x = self.model.pval()
                y0 = self.model.ival()
                obj_val = self.model.problem.objective(None, x, y0)
                writer.writerow(["original", f"{obj_val:8.3e}"])
            n_file = self.get_executable()
            for paramset in sorted(n_file):
                optimized = self.load_param(paramset)
                obj_val = self.model.problem.objective(None, *optimized)
                writer.writerow([f"{paramset:d}", f"{obj_val:8.3e}"])

    def trace_obj(
        self,
        *,
        config: Optional[dict] = None,
        xlabel: str = "Iteration",
        ylabel: str = "Objective function value",
        xticks: Optional[list] = None,
        yticks: Optional[list] = None,
    ) -> None:
        """
        Visualize objective function traces for different optimization runs.

        Parameters
        ----------
        config : dict, optional
            A dictionary object for setting `matplotlib.rcParams`.
        xlabel: str (default: "Iteration")
            The label for the x-axis.
        ylabel: str (default: "Objective function value")
            The label for the x-axis.
        xticks: list, optional
            The list of xtick locations.
        yticks: list, optional
            The list of ytick locations.

        Examples
        --------
        >>> from biomass import Model, OptimizationResults
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> model = Model(Nakakuki_Cell_2010.__package__).create()
        >>> res = OptimizationResults(model)
        >>> res.trace_obj()

        Notes
        -----
        Output:

        * optimization_results/obj_func_traces.pdf

        """
        n_file = self.get_executable()
        # matplotlib
        if config is None:
            config = {}
        config.setdefault("font.size", 15)
        config.setdefault("axes.linewidth", 1.5)
        config.setdefault("xtick.major.width", 1.5)
        config.setdefault("ytick.major.width", 1.5)
        config.setdefault("lines.linewidth", 1.5)
        config.setdefault("savefig.bbox", "tight")
        config.setdefault("savefig.format", "pdf")
        plt.rcParams.update(config)
        plt.gca().spines["right"].set_visible(False)
        plt.gca().spines["top"].set_visible(False)
        # ---
        for paramset in n_file:
            with open(
                os.path.join(self.model.path, "out", f"{paramset:d}", "optimization.log"),
                mode="r",
            ) as f:
                traces = f.readlines()
            iters = []
            obj_val = []
            for line in traces:
                if line.startswith("Generation"):
                    iters.append(line.lstrip("Generation").split(":")[0])
                    obj_val.append(line.split("=")[-1].strip())
            plt.plot([int(num) - 1 for num in iters], [float(val) for val in obj_val])
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        if xticks is not None:
            plt.xticks(xticks)
        if yticks is not None:
            plt.yticks(yticks)
        plt.savefig(os.path.join(self.model.path, "optimization_results", "obj_func_traces"))
        plt.close()



import math
import operator
import sys
import time
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
from scipy.integrate import ode, solve_ivp
from scipy.integrate._ivp.ivp import OdeResult

__all__ = ["solve_ode", "get_steady_state"]


def solve_ode(
    diffeq: Callable,
    y0: Union[list, np.ndarray],
    t: Union[range, List[int]],
    f_params: Tuple[float, ...],
    method: str = "BDF",
    options: Optional[dict] = None,
) -> Optional[OdeResult]:
    """
    Solve a system of ordinary differential equations.

    Parameters
    ----------
    diffeq : callable f(t, y, *x)
        Right-hand side of the differential equation.
    y0 : array
        Initial condition on y (can be a vector).
    t : array
        A sequence of time points for which to solve for y.
    f_params : tuple
        Model parameters.
    method : str (default: "BDF")
        Integration method to use.
    options : dict, optional
        Options passed to a chosen solver.

    Returns
    -------
    sol : OdeResult
        Represents the solution of ODE.
    """
    if options is None:
        options = {}
    options.setdefault("rtol", 1e-8)
    options.setdefault("atol", 1e-8)
    try:
        sol = solve_ivp(
            diffeq,
            (t[0], t[-1]),
            y0,
            method=method,
            t_eval=t,
            args=f_params,
            **options,
        )
        return sol if sol.success else None
    except ValueError:
        return None


def get_steady_state(
    diffeq: Callable,
    y0: list,
    f_params: tuple,
    *,
    dt: float = 1,
    atol: float = 1e-8,
    rtol: float = 1e-8,
    eps: float = 1e-6,
    maximum_wait_time: Union[int, float] = 60.0,
) -> List[float]:
    """
    Use an ODE solver to find the steady state for the untreated condition.

    Parameters
    ----------
    diffeq : callable f(t, y, *x)
        Right-hand side of the differential equation.
    y0 : array
        Initial condition on y (can be a vector).
    f_params : tuple
        Model parameters.
    dt : float (default: 1.0)
        The step size used to calculate the steady state.
    atol : float (default: 1e-8)
        Absolute tolerance for solution.
    rtol : float (default: 1e-8)
        Relative tolerance for solution.
    eps : float (default: 1e-6)
        Run until a time t for which the maximal absolute value of the
        regularized relative derivative was smaller than eps.
    maximum_wait_time : int or float (default: 60.0 = 1 min.)
        The longest time a user can wait for the system to reach the steady state.

    Returns
    -------
    steady_state : List[float]
        Steady state concentrations of all species.
        Return an empty list if simulation failed.
    """

    sol = ode(lambda t, y, f_args: diffeq(t, y, *f_args))
    sol.set_integrator(
        "zvode",
        method="bdf",
        with_jacobian=True,
        atol=atol,
        rtol=rtol,
    )
    sol.set_initial_value(y0, 0)
    sol.set_f_params(f_params)

    ys = [y0]
    start = time.time()
    while sol.successful():
        sol.integrate(sol.t + dt)
        if (
            np.iscomplex(np.real_if_close(sol.y, tol=1)).any()
            or (time.time() - start) > maximum_wait_time
        ):
            return []
        elif (
            np.max(
                np.abs(
                    np.real(list(map(operator.sub, sol.y, ys[-1])))
                    / np.real([(yi + eps) for yi in sol.y])
                )
            )
            < eps
        ):
            break
        else:
            ys.append(sol.y)
    steady_state = np.real(sol.y).tolist() if sol.successful() else []
    for i, val in enumerate(steady_state):
        if math.fabs(val) < sys.float_info.epsilon:
            steady_state[i] = 0.0
    return steady_state



import sys
from dataclasses import dataclass, field
from math import fabs, isnan, log
from typing import Callable, Dict, List, Union

import numpy as np
from scipy.integrate import simpson


@dataclass
class SignalingMetric(object):
    """
    Signaling metric used for sensitivity analysis.

    Attributes
    ----------
    quantification : Dict[str, Callable[[np.ndarray], Union[int, float]]]
        Functions to quantify signaling metrics.
    """

    quantification: Dict[str, Callable[[np.ndarray], Union[int, float]]] = field(
        default_factory=lambda: dict(
            maximum=np.max,
            minimum=np.min,
            integral=simpson,
        ),
        init=False,
    )


def dlnyi_dlnxj(
    signaling_metric: np.ndarray,
    n_file: List[int],
    perturbed_idx: List[int],
    observables: List[str],
    conditions: List[str],
    rate: float,
) -> np.ndarray:
    """
    Numerical computation of sensitivities using finite difference approximations
    with 1% changes in the reaction rates, parameter values or non-zero initial values.

    Parameters
    ----------
    signaling_metric : numpy array
        Signaling metric

    n_file : list of integers
        Optimized parameter sets in out/

    perturbed_idx : list of integers
        Indices of rate equations or non-zero initial values.

    observables : list of strings
        observables in observable.py

    conditions : list of strings
        Experimental conditions.

    rate : float ~ 1
        1.01 for 1% change.

    Returns
    -------
    sensitivity_coefficients: numpy array

    """
    sensitivity_coefficients = np.empty(
        (len(n_file), len(perturbed_idx), len(observables), len(conditions))
    )
    for i, _ in enumerate(n_file):
        for j, _ in enumerate(perturbed_idx):
            for k, _ in enumerate(observables):
                for l, _ in enumerate(conditions):
                    if isnan(signaling_metric[i, j, k, l]):
                        sensitivity_coefficients[i, j, k, l] = np.nan
                    elif (
                        fabs(signaling_metric[i, -1, k, l]) < sys.float_info.epsilon
                        or fabs(signaling_metric[i, j, k, l] - signaling_metric[i, -1, k, l])
                        < sys.float_info.epsilon
                        or (signaling_metric[i, j, k, l] / signaling_metric[i, -1, k, l]) <= 0
                    ):
                        # 1. Signaling metric before adding perturbation is zero
                        # 2. Absolute change caused by perturbation is too small
                        # 3. Antilogarithm <= 0
                        sensitivity_coefficients[i, j, k, l] = 0.0
                    else:
                        sensitivity_coefficients[i, j, k, l] = log(
                            signaling_metric[i, j, k, l] / signaling_metric[i, -1, k, l]
                        ) / log(rate)

    return sensitivity_coefficients



from .name2idx import C, V


class DifferentialEquation(object):
    """Kinetic equations comprising the computational model of the MAPK cascade."""

    def __init__(self, perturbation):
        super(DifferentialEquation, self).__init__()
        self.perturbation = perturbation

    def diffeq(self, t, y, *x):
        # Rate equation
        v = {}
        v[1] = (
            x[C.V1]
            * y[V.MKKK]
            / ((1 + (y[V.MAPK_PP] / x[C.KI]) ** x[C.n]) * (x[C.K1] + y[V.MKKK]))
        )
        v[2] = x[C.V2] * y[V.MKKK_P] / (x[C.K2] + y[V.MKKK_P])
        v[3] = x[C.k3] * y[V.MKKK_P] * y[V.MKK] / (x[C.K3] + y[V.MKK])
        v[4] = x[C.k4] * y[V.MKKK_P] * y[V.MKK_P] / (x[C.K3] + y[V.MKK_P])
        v[5] = x[C.V5] * y[V.MKK_PP] / (x[C.K5] + y[V.MKK_PP])
        v[6] = x[C.V6] * y[V.MKK_P] / (x[C.K6] + y[V.MKK_P])
        v[7] = x[C.k7] * y[V.MKK_PP] * y[V.MAPK] / (x[C.K7] + y[V.MAPK])
        v[8] = x[C.k8] * y[V.MKK_PP] * y[V.MAPK_P] / (x[C.K8] + y[V.MAPK_P])
        v[9] = x[C.V9] * y[V.MAPK_PP] / (x[C.K9] + y[V.MAPK_PP])
        v[10] = x[C.V10] * y[V.MAPK_P] / (x[C.K10] + y[V.MAPK_P])

        if self.perturbation:
            for i, dv in self.perturbation.items():
                v[i] = v[i] * dv

        dydt = [0] * V.NUM
        dydt[V.MKKK] = v[2] - v[1]
        dydt[V.MKKK_P] = v[1] - v[2]
        dydt[V.MKK] = v[6] - v[3]
        dydt[V.MKK_P] = v[3] + v[5] - v[4] - v[6]
        dydt[V.MKK_PP] = v[4] - v[5]
        dydt[V.MAPK] = v[10] - v[7]
        dydt[V.MAPK_P] = v[7] + v[9] - v[8] - v[10]
        dydt[V.MAPK_PP] = v[8] - v[9]

        return dydt


def param_values():
    # Parameter values
    x = [0] * C.NUM
    x[C.V1] = 2.5
    x[C.n] = 1
    x[C.KI] = 9
    x[C.K1] = 10
    x[C.V2] = 0.25
    x[C.K2] = 8
    x[C.k3] = 0.025
    x[C.K3] = 15
    x[C.k4] = 0.025
    x[C.K4] = 15
    x[C.V5] = 0.75
    x[C.K5] = 15
    x[C.V6] = 0.75
    x[C.K6] = 15
    x[C.k7] = 0.025
    x[C.K7] = 15
    x[C.k8] = 0.025
    x[C.K8] = 15
    x[C.V9] = 0.5
    x[C.K9] = 15
    x[C.V10] = 0.5
    x[C.K10] = 15

    return x


def initial_values():
    # Value of the initial condition
    y0 = [0] * V.NUM

    y0[V.MKKK] = 90
    y0[V.MKKK_P] = 10
    y0[V.MKK] = 280
    y0[V.MKK_P] = 10
    y0[V.MKK_PP] = 10
    y0[V.MAPK] = 280
    y0[V.MAPK_P] = 10
    y0[V.MAPK_PP] = 10

    return y0



"""
References
----------
- DIDC:
    Kimura, S. & Konagaya, A. A Genetic Algorithm with Distance
    Independent Diversity Control for High Dimensional Function
    Optimization. J. Japanese Soc. Artif. Intell. 18, 193–202 (2003).

- AGLSDC:
    Kimura S, Nakakuki T, Kirita S, Okada M. AGLSDC: A Genetic Local Search
    Suitable for Parallel Computation. SICE J Control Meas Syst Integr 2012;
    4: 105–113.
"""

from dataclasses import dataclass, field
from math import isnan
from typing import Callable

import numpy as np
from scipy.optimize import differential_evolution, minimize


@dataclass(frozen=True)
class OptimizeStep(object):
    obj_func: Callable[[np.ndarray], float]
    n_population: int
    n_gene: int
    n_children: int
    maxiter: int
    workers: int
    n_children_for_endx: int = field(default=10, init=False)

    def _xover(self, parents: np.ndarray) -> np.ndarray:
        """Extended Normal Distribution Xover"""
        ALPHA = (1.0 - 2 * 0.35 ** 2) ** 0.5 / 2.0
        BETA = 0.35 / (self.n_gene - 1) ** 0.5

        child = np.empty(self.n_gene + 1)

        t1 = (parents[1, : self.n_gene] - parents[0, : self.n_gene]) / 2.0
        t2 = np.random.normal(scale=ALPHA) * (
            parents[1, : self.n_gene] - parents[0, : self.n_gene]
        )
        t3 = np.sum(
            np.random.normal(scale=BETA, size=self.n_gene)[:, np.newaxis]
            * (
                parents[2:, : self.n_gene]
                - (np.sum(parents[2:, : self.n_gene], axis=0) / self.n_gene)
            ),
            axis=0,
        )
        child[: self.n_gene] = t1 + t2 + t3
        child[: self.n_gene] = np.clip(child[: self.n_gene], 0.0, 1.0)
        child[-1] = 1e12  # assigns the worst objective value to the children.

        return child

    def converging(self, ip: np.ndarray, population: np.ndarray) -> np.ndarray:
        children = np.empty((self.n_children_for_endx, self.n_gene + 1))

        for i in range(self.n_children_for_endx):
            ip[2:] = np.random.choice(self.n_population, self.n_gene, replace=False)
            children[i, :] = self._xover(population[ip, :])

        family = np.empty((self.n_children_for_endx + 2, self.n_gene + 1))
        family[: self.n_children_for_endx, :] = children
        family[-2, :] = population[ip[0], :]
        family[-1, :] = population[ip[1], :]

        family = family[np.argsort(family[:, -1]), :]
        # Best, either of parents
        population[ip[0], :] = family[0, :]
        # Random
        random_child_idx = np.random.randint(
            low=1, high=self.n_children_for_endx + 2, dtype=np.int
        )
        population[ip[1], :] = family[random_child_idx, :]

        if 1e12 <= population[ip[1], -1] or isnan(population[ip[1], -1]):
            population[ip[1], -1] = self.obj_func(population[ip[1], : self.n_gene])

        population = population[np.argsort(population[:, -1]), :]

        return population

    def _mutation(self, parents: np.ndarray) -> np.ndarray:
        """Normal Distribution Mutation"""
        GAMMA = 0.35 / self.n_gene ** 0.5

        child = np.empty(self.n_gene + 1)

        t2 = np.sum(
            np.random.normal(scale=GAMMA, size=self.n_gene + 1)[:, np.newaxis]
            * (
                parents[1:, : self.n_gene]
                - (np.sum(parents[1:, : self.n_gene], axis=0) / (self.n_gene + 1))
            ),
            axis=0,
        )
        child[: self.n_gene] = parents[0, : self.n_gene] + t2
        child[: self.n_gene] = np.clip(child[: self.n_gene], 0.0, 1.0)
        child[-1] = self.obj_func(child[: self.n_gene])

        return child

    def local_search(
        self,
        ip: np.ndarray,
        population: np.ndarray,
        method: str,
    ) -> np.ndarray:
        """
        Apply the local search method to the best individual in a family
        consisting of the two parents, i.e., p1 and p2, and their children.
        """
        if method == "mutation":
            idx = [True] * self.n_population
            idx[ip[0]] = False
            children = np.empty((self.n_children, self.n_gene + 1))
            for i in range(self.n_children):
                ip[1:] = np.random.choice(
                    np.arange(self.n_population)[idx], self.n_gene + 1, replace=False
                )
                children[i, :] = self._mutation(population[ip, :])
            family = np.empty((self.n_children + 1, self.n_gene + 1))
            family[: self.n_children, :] = children
            family[-1, :] = population[ip[0], :]
            family = family[np.argsort(family[:, -1]), :]
            population[ip[0], :] = family[0, :]  # Elite
        elif method == "powell":
            lower = np.min(population[:, : self.n_gene], axis=0)
            upper = np.max(population[:, : self.n_gene], axis=0)
            direc = np.identity(self.n_gene) * 0.3 * (upper - lower)
            res = minimize(
                self.obj_func,
                population[ip[0], : self.n_gene],
                method="Powell",
                bounds=tuple(zip(lower, upper)),
                callback=lambda xk: True
                if self.obj_func(xk) < self.obj_func(population[ip[0], : self.n_gene])
                else False,
                options={
                    "xtol": 1.0,
                    "ftol": 1.0,
                    "maxiter": self.maxiter,
                    "maxfev": 100 * self.n_gene,
                    "direc": direc,
                },
            )
            if res.fun < self.obj_func(population[ip[0], : self.n_gene]):
                population[ip[0], : self.n_gene] = res.x
                population[ip[0], -1] = res.fun
        elif method == "de":
            lower = np.min(population[:, : self.n_gene], axis=0)
            upper = np.max(population[:, : self.n_gene], axis=0)
            res = differential_evolution(
                self.obj_func,
                tuple(zip(lower, upper)),
                strategy="best2bin",
                mutation=(0.0, 0.3),
                recombination=0.9,
                maxiter=self.maxiter,
                popsize=1,
                polish=False,
                init=population[ip, : self.n_gene],
                updating="immediate" if self.workers == 1 else "deferred",
                workers=self.workers,
            )
            if res.fun < self.obj_func(population[ip[0], : self.n_gene]):
                population[ip[0], : self.n_gene] = res.x
                population[ip[0], -1] = res.fun

        population = population[np.argsort(population[:, -1]), :]

        return population



from matplotlib import pyplot as plt

from biomass.plotting import *

from .observable import Observable


class Visualization(Observable):
    """
    Plotting parameters for customizing figure properties.

    Attributes
    ----------
    cm : matplotlib.colors.ListedColormap (default: `plt.cm.get_cmap('tab10')`)
        Choosing colormaps for `cmap`.
    single_observable_options : list of SingleObservable
        Visualization options for time-course simulation (single-observable).
    multiple_observables_options : MultipleObservables
        Visualization options for time-course simulation (multi-observables).
    sensitivity_options : SensitivityOptions
        Visualization options for sensitivity analysis results.
    """

    def __init__(self):
        super().__init__()

        self.cm = plt.cm.get_cmap("tab10")
        self.single_observable_options = [
            SingleObservable(self.cm, obs_name) for obs_name in self.obs_names
        ]
        self.multiple_observables_options = MultipleObservables(self.cm)
        self.sensitivity_options = SensitivityOptions(self.cm)

    def get_single_observable_options(self):
        
        for i, _ in enumerate(self.obs_names):
            self.single_observable_options[i].divided_by = 60  # sec. -> min.
            self.single_observable_options[i].xlim = (-5, 95)
            self.single_observable_options[i].xticks = [0, 30, 60, 90]
            self.single_observable_options[i].xlabel = 'Time (min)'
            self.single_observable_options[i].ylim = (-0.1, 1.3)
            self.single_observable_options[i].yticks = [0.0, 0.3, 0.6, 0.9, 1.2]
            self.single_observable_options[i].cmap = ['mediumblue', 'red']
            self.single_observable_options[i].shape = ['D', 's']
            self.single_observable_options[i].dont_show = []

        self.single_observable_options[
            self.obs_names.index('Phosphorylated_MEKc')
        ].ylabel = 'Phosphorylated MEK\n(cytoplasm)'

        self.single_observable_options[
            self.obs_names.index('Phosphorylated_ERKc')
        ].ylabel = 'Phosphorylated ERK\n(cytoplasm)'

        self.single_observable_options[
            self.obs_names.index('Phosphorylated_RSKw')
        ].ylabel = 'Phosphorylated RSK\n(whole cell)'

        self.single_observable_options[
            self.obs_names.index('Phosphorylated_CREBw')
        ].ylabel = 'Phosphorylated CREB\n(whole cell)'

        self.single_observable_options[
            self.obs_names.index('dusp_mRNA')
        ].ylabel = r'$\it{dusp}$'+' mRNA\nexpression'

        self.single_observable_options[
            self.obs_names.index('cfos_mRNA')
        ].ylabel = r'$\it{c}$'+'-'+r'$\it{fos}$'+' mRNA\nexpression'

        self.single_observable_options[
            self.obs_names.index('cFos_Protein')
        ].ylabel = 'c-Fos Protein\nexpression'

        self.single_observable_options[
            self.obs_names.index('Phosphorylated_cFos')
        ].ylabel = 'Phosphorylated c-Fos\nProtein expression'
        
        return self.single_observable_options

    def get_multiple_observables_options(self):
        """
        self.multiple_observables_options.observables = [
            'Phosphorylated_ERKc',
            'Phosphorylated_CREBw',
            'Phosphorylated_cFos',
            'dusp_mRNA'
        ]
        self.multiple_observables_options.condition = 'EGF'
        self.multiple_observables_options.xlim = (-5, 95)
        self.multiple_observables_options.xticks = [0, 30, 60, 90]
        self.multiple_observables_options.xlabel = 'Time (min)'
        self.multiple_observables_options.ylim = (-0.1, 1.3)
        self.multiple_observables_options.yticks = [0.0, 0.3, 0.6, 0.9, 1.2]
        self.multiple_observables_options.ylabel = 'Intensity (a.u.)'
        """
        return self.multiple_observables_options

    def get_sensitivity_options(self):

        return self.sensitivity_options
        
    @staticmethod
    def set_timecourse_rcParams():
        """figure/simulation"""
        plt.rcParams["font.size"] = 20
        plt.rcParams["axes.linewidth"] = 1.5
        plt.rcParams["xtick.major.width"] = 1.5
        plt.rcParams["ytick.major.width"] = 1.5
        plt.rcParams["lines.linewidth"] = 1.8
        plt.rcParams["lines.markersize"] = 12
        plt.rcParams["savefig.bbox"] = "tight"
        # plt.rcParams["savefig.format"] = "pdf"
        # plt.rcParams["savefig.transparent"] = False
        # plt.rcParams["savefig.dpi"] = 1200
        # plt.rcParams['font.family'] = 'Arial'
        # plt.rcParams['mathtext.fontset'] = 'custom'
        # plt.rcParams['mathtext.it'] = 'Arial:italic'

    @staticmethod
    def set_sensitivity_rcParams():
        """figure/sensitivity"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.2
        plt.rcParams["xtick.major.width"] = 1.2
        plt.rcParams["ytick.major.width"] = 1.2
        plt.rcParams["savefig.bbox"] = "tight"
        # plt.rcParams["savefig.format"] = "pdf"
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def convert_species_name(name):
        """figure/sensitivity/initial_condition
        - Sensitivity for species with nonzero initial conditions
        """
        """
        if name == 'ERKc':
            return 'ERK (cytoplasm)'
        elif name == 'RSKc':
            return 'RSK (cytoplasm)'
        elif name == 'CREBn':
            return 'CREB (nucleus)'
        elif name == 'Elk1n':
            return 'Elk1 (nucleus)'
        """
        return name



"""BioMASS core functions"""
import multiprocessing
import os
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional, Union

import numpy as np

from .analysis import InitialConditionSensitivity, ParameterSensitivity, ReactionSensitivity
from .dynamics import SignalingSystems
from .estimation import GeneticAlgorithmContinue, GeneticAlgorithmInit
from .exec_model import ModelObject

__all__ = ["Model", "optimize", "optimize_continue", "run_simulation", "run_analysis"]


@dataclass
class Model(object):
    """
    The BioMASS model object.

    Attributes
    ----------
    pkg_name: str
        Path (dot-sepalated) to a biomass model directory.
        Use '__package__'.
    """

    pkg_name: str

    def _load_model(self) -> Any:
        try:
            biomass_model = import_module(self.pkg_name)
            return biomass_model
        except ImportError:
            p = Path(self.pkg_name.replace(".", os.sep))
            print(f"cannot import '{p.name}' from '{p.parent}'.")

    def create(self, show_info: bool = False) -> ModelObject:
        """
        Build a biomass model.

        Parameters
        ----------
        show_info : bool (default: False)
            Set to `True` to print the information related to model size.

        Examples
        --------
        >>> from biomass import Model
        >>> import your_model
        >>> model = Model(your_model.__package__).create()
        """
        model = ModelObject(self.pkg_name.replace(".", os.sep), self._load_model())
        if model.problem.normalization:
            for obs_name in model.observables:
                if (
                    isinstance(model.problem.normalization[obs_name]["timepoint"], int)
                    and not model.problem.t[0]
                    <= model.problem.normalization[obs_name]["timepoint"]
                    <= model.problem.t[-1]
                ):
                    raise ValueError("Normalization timepoint must lie within problem.t.")
                if not model.problem.normalization[obs_name]["condition"]:
                    model.problem.normalization[obs_name]["condition"] = model.problem.conditions
                else:
                    for c in model.problem.normalization[obs_name]["condition"]:
                        if c not in model.problem.conditions:
                            raise ValueError(
                                f"Normalization condition '{c}' is not defined in problem.conditions."
                            )
        if show_info:
            model_name = Path(model.path).name
            print(
                f"{model_name} information\n" + ("-" * len(model_name)) + "------------\n"
                f"{len(model.species):d} species\n"
                f"{len(model.parameters):d} parameters, "
                f"of which {len(model.problem.idx_params):d} to be estimated"
            )
        return model


def _check_optional_arguments(
    x_id: Union[int, Iterable[int]],
    options: Optional[dict],
) -> None:
    if options is None:
        pass
    elif isinstance(options, dict):
        if options["local_search_method"].lower() not in ["mutation", "powell", "de"]:
            raise ValueError(
                f"'{options['local_search_method']}': "
                "Invalid local_search_method. Should be one of ['mutation', 'Powell', 'DE']"
            )
        elif (
            not isinstance(x_id, int)
            and options["local_search_method"].lower() == "de"
            and options["workers"] != 1
        ):
            raise AssertionError(
                "daemonic processes are not allowed to have children. Set options['workers'] to 1."
            )
    else:
        raise TypeError("options must be dict or None.")


def optimize(
    model: ModelObject,
    x_id: Union[int, Iterable[int]],
    options: Optional[dict] = None,
) -> None:
    """
    Estimate model parameters from experimental data.

    Parameters
    ----------
    model : ModelObject
        Model for parameter estimation.

    x_id : int or Iterable[int]
        Index (indices) of parameter set to estimate.

    options : dict, optional
        * popsize : int (default: 5)
            A multiplier for setting the total population size.
            The population has `popsize` * len(`search_param`) individuals.

        * max_generation : int (default: 10000)
            Stop optimization if Generation > `max_generation`.

        * initial_threshold : float (default: 1e12)
            Threshold on objective function value used to generate initial population.
            Default value is 1e12 (numerically solvable).

        * allowable_error : float (default: 0.0)
            Stop optimization if Best Fitness <= `allowable_error`.

        * local_search_method : str (default: 'mutation')
            Method used in local search. Should be one of:

            * 'mutation' : NDM/MGG

            * 'Powell' : Modified Powell method

            * 'DE' : Differential Evolution (strategy: `best2bin`)

        * n_children : int (default: 1000)
            (method='mutation') The number of children generated in NDM/MGG.

        * maxiter : int (default: 10)
            (method='Powell' or 'DE') The maximum number of iterations
            over which the entire population is evolved.

        * workers : int (default: -1 if isinstance(x_id, `int`) else 1)
            (method='DE') The population is subdivided into workers sections and
            evaluated in parallel (uses multiprocessing.Pool). Supply -1 to use
            all available CPU cores. Set workers to 1 when searching multiple
            parameter sets simultaneously.

        * overwrite : bool (default: False)
            If `True`, the out/n folder will be overwritten.

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import Model, optimize
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()
    >>> optimize(
    ...     model, x_id=range(1, 11),
    ...     options={
    ...         'max_generation': 10000,
    ...         'allowable_error': 0.5
    ...     }
    ... )

    """
    if options is None:
        options = {}
    options.setdefault("popsize", 5)
    options.setdefault("max_generation", 10000)
    options.setdefault("initial_threshold", 1e12)
    options.setdefault("allowable_error", 0.0)
    options.setdefault("local_search_method", "mutation")
    options.setdefault("n_children", 1000)
    options.setdefault("maxiter", 10)
    options.setdefault("workers", -1 if isinstance(x_id, int) else 1)
    options.setdefault("overwrite", False)

    _check_optional_arguments(x_id, options)

    ga_init = GeneticAlgorithmInit(model, **options)
    if isinstance(x_id, int):
        ga_init.run(x_id)
    elif all([isinstance(i, int) for i in x_id]):
        n_proc = max(1, multiprocessing.cpu_count() - 1)
        with multiprocessing.Pool(processes=n_proc) as p:
            for _ in p.imap_unordered(ga_init.run, x_id):
                pass
    else:
        raise TypeError("`x_id` type must be either int or Iterable[int].")


def optimize_continue(
    model: ModelObject,
    x_id: Union[int, Iterable[int]],
    options: Optional[dict] = None,
) -> None:
    """
    Continue running optimization from where you stopped in the last parameter search.

    Parameters
    ----------
    model : ModelObject
        Model for parameter estimation.

    x_id : int or Iterable[int]
        Index (indices) of parameter set to estimate.

    options : dict, optional
        * popsize : int (default: 5)
            A multiplier for setting the total population size.
            The population has `popsize` * len(`search_param`) individuals.

        * max_generation : int (default: 15000)
            Stop optimization if Generation > `max_generation`.

        * initial_threshold : float (default: 1e12)
            Threshold on objective function value used to generate initial population.
            Default value is 1e12 (numerically solvable).

        * allowable_error : float (default: 0.0)
            Stop optimization if Best Fitness <= `allowable_error`.

        * local_search_method : str (default: 'mutation')
            Method used in local search. Should be one of:

            * 'mutation' : NDM/MGG

            * 'Powell' : Modified Powell method

            * 'DE' : Differential Evolution (strategy: `best2bin`)

        * n_children : int (default: 1000)
            (method='mutation') The number of children generated in NDM/MGG.

        * maxiter : int (default: 10)
            (method='Powell' or 'DE') The maximum number of iterations
            over which the entire population is evolved.

        * workers : int (default: -1 if isinstance(x_id, `int`) else 1)
            (method='DE') The population is subdivided into workers sections and
            evaluated in parallel (uses multiprocessing.Pool). Supply -1 to use
            all available CPU cores. Set workers to 1 when searching multiple
            parameter sets simultaneously.

        * p0_bounds : list of floats (default: [0.1, 10.0])
            Generate initial population using best parameter values in the last
            parameter search.

            * `lower_bound` = po_bounds[0] * `best_parameter_value`
            * `upper_bound` = p0_bounds[1] * `best_parameter_value`

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import Model, optimize_continue
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()
    >>> optimize_continue(
    ...     model, x_id=range(1, 11),
    ...     options={
    ...         'max_generation': 20000,
    ...         'allowable_error': 0.5
    ...     }
    ... )

    """
    if options is None:
        options = {}
    options.setdefault("popsize", 5)
    options.setdefault("max_generation", 15000)
    options.setdefault("initial_threshold", 1e12)
    options.setdefault("allowable_error", 0.0)
    options.setdefault("local_search_method", "mutation")
    options.setdefault("n_children", 1000)
    options.setdefault("maxiter", 10)
    options.setdefault("workers", -1 if isinstance(x_id, int) else 1)
    options.setdefault("p0_bounds", [0.1, 10.0])

    _check_optional_arguments(x_id, options)

    ga_continue = GeneticAlgorithmContinue(model, **options)
    if isinstance(x_id, int):
        ga_continue.run(x_id)
    elif all([isinstance(i, int) for i in x_id]):
        n_proc = max(1, multiprocessing.cpu_count() - 1)
        with multiprocessing.Pool(processes=n_proc) as p:
            for _ in p.imap_unordered(ga_continue.run, x_id):
                pass
    else:
        raise TypeError("`x_id` type must be either int or Iterable[int].")


def run_simulation(
    model: ModelObject,
    *,
    viz_type: str = "original",
    show_all: bool = False,
    stdev: bool = False,
) -> None:
    """
    Simulate ODE model with estimated parameter values.

    Parameters
    ----------
    model : ModelObject
        Model for simulation.

    viz_type : str
        * 'average':
            The average of simulation results with parameter sets in "out/".
        * 'best':
            The best simulation result in "out/", simulation with
            "best_fit_param".
        * 'original':
            Simulation with the default parameters and initial values
            defined in "set_model.py".
        * 'n(=1,2,...)':
            Use the parameter set in "out/n/".
        * 'experiment'
            Draw the experimental data written in observable.py without
            simulation results.

    show_all : bool
        Whether to show all simulation results.

    stdev : bool
        If True, the standard deviation of simulated values will be shown
        (only available for 'average' visualization type).

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import Model, run_simulation
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()
    >>> run_simulation(
    ...     model,
    ...     viz_type='average',
    ...     show_all=False,
    ...     stdev=True,
    ... )

    """
    if viz_type not in ["best", "average", "original", "experiment"] and not viz_type.isdecimal():
        raise ValueError(
            "Available viz_type are: 'best','average','original','experiment','n(=1, 2, ...)'"
        )

    SignalingSystems(model).simulate_all(
        viz_type=viz_type,
        show_all=show_all,
        stdev=stdev,
    )


def run_analysis(
    model: ModelObject,
    *,
    target: str,
    metric: str = "integral",
    create_metrics: Optional[Dict[str, Callable[[np.ndarray], Union[int, float]]]] = None,
    style: str = "barplot",
    options: Optional[dict] = None,
) -> None:
    """
    Employ sensitivity analysis to identify critical parameters, species or
    reactions in the complex biological network.

    The sensitivity S(y,x) was calculated according to the following equation:
    S(y,x) = d ln(yi) / d ln (xj), where yi is the signaling metric and xj is
    each nonzero species, parameter value or reaction rate.

    Parameters
    ---------
    model : ModelObject
        Model for sensitivity analysis.

    target : str
        * 'reaction'
        * 'initial_condition'
        * 'parameter'

    metric : str (default: 'integral')
        A word to specify the signaling metric.

    create_metrics : Dict[str, Callable[[np.ndarray], Union[int, float]]], optional
        Create user-defined signaling metrics.

    style : str (default: 'barplot')
        * 'barplot'
        * 'heatmap'

    options : dict, optional
        * show_indices : bool (default: True)
            (target == 'reaction') Set to `True` to put reaction index on each bar.

        * excluded_params : list of strings
            (target == 'parameter') List of parameters which are not used for analysis.

        * excluded_initials : list of strings
            (target == 'initial_condition') List of species which are not used for analysis.
        
        * overwrite : bool (default: False)
            If `True`, the sensitivity_coefficients/{target}/{metric}.npy file will be overwritten.

    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import Model, run_analysis
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()

    Parameters

    >>> run_analysis(
    ...     model,
    ...     target='parameter',
    ...     options = {
    ...         'excluded_params': [
    ...             'a', 'Vn', 'Vc', 'Ligand', 'EGF', 'HRG', 'no_ligand'
    ...         ]
    ...     }
    ... )

    Initial condition

    >>> run_analysis(model, target='initial_condition')

    Reaction

    >>> run_analysis(model, target='reaction')

    """

    if options is None:
        options = {}
    options.setdefault("show_indices", True)
    options.setdefault("excluded_params", [])
    options.setdefault("excluded_initials", [])
    options.setdefault("overwrite", False)

    if target == "reaction":
        ReactionSensitivity(model, create_metrics).analyze(
            metric=metric,
            style=style,
            options=options,
        )
    elif target == "parameter":
        ParameterSensitivity(model, create_metrics).analyze(
            metric=metric,
            style=style,
            options=options,
        )
    elif target == "initial_condition":
        InitialConditionSensitivity(model, create_metrics).analyze(
            metric=metric,
            style=style,
            options=options,
        )
    else:
        raise ValueError(
            "Available targets are: '{}".format(
                "', '".join(["reaction", "parameter", "initial_condition"]) + "'."
            )
        )



import os
import sys
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ..exec_model import ExecModel, ModelObject
from ..plotting import SensitivityOptions
from .util import SignalingMetric, dlnyi_dlnxj


@dataclass
class InitialConditionSensitivity(ExecModel, SignalingMetric):
    """Sensitivity for species with nonzero initial conditions"""

    model: ModelObject
    create_metrics: Optional[Dict[str, Callable[[np.ndarray], Union[int, float]]]]

    def __post_init__(self) -> None:
        self._plotting: SensitivityOptions = self.model.viz.get_sensitivity_options()
        self._coefficients: Callable[[str], str] = lambda metric: os.path.join(
            self.model.path,
            "sensitivity_coefficients",
            "initial_condition",
            f"{metric}.npy",
        )
        self._path_to_figs: Callable[[str], str] = lambda metric: os.path.join(
            self.model.path,
            "figure",
            "sensitivity",
            "initial_condition",
            f"{metric}",
        )
        if self.create_metrics is not None:
            for name, function in self.create_metrics.items():
                self.quantification[name] = function

    def _get_nonzero_indices(self, excluded_initials: List[str]) -> List[int]:
        nonzero_indices = []
        y0 = self.model.ival()
        for i, val in enumerate(y0):
            if self.model.species[i] not in excluded_initials and val != 0.0:
                nonzero_indices.append(i)
        if not nonzero_indices:
            raise ValueError("No nonzero initial conditions.")

        return nonzero_indices

    def _calc_sensitivity_coefficients(
        self,
        metric: str,
        nonzero_indices: List[int],
    ) -> np.ndarray:
        """Calculating Sensitivity Coefficients

        Parameters
        ----------
        metric : str
            The signaling metric used for sensitivity analysis.

        nonzero_indices : list of int
            List of species index with nonzero initial conditions.

        Returns
        -------
        sensitivity_coefficients : numpy array

        """

        rate = 1.01  # 1% change
        n_file = self.get_executable()

        signaling_metric = np.full(
            (
                len(n_file),
                len(nonzero_indices) + 1,
                len(self.model.observables),
                len(self.model.problem.conditions),
            ),
            np.nan,
        )
        for i, nth_paramset in enumerate(n_file):
            optimized = self.load_param(nth_paramset)
            for j, idx in enumerate(nonzero_indices):
                y0 = optimized.initials[:]
                y0[idx] = optimized.initials[idx] * rate
                if self.model.problem.simulate(optimized.params, y0) is None:
                    for k, _ in enumerate(self.model.observables):
                        for l, _ in enumerate(self.model.problem.conditions):
                            signaling_metric[i, j, k, l] = self.quantification[metric](
                                self.model.problem.simulations[k, :, l]
                            )
                sys.stdout.write(
                    "\r{:d} / {:d}".format(
                        i * len(nonzero_indices) + j + 1,
                        len(n_file) * len(nonzero_indices),
                    )
                )
            # Signaling metric without perturbation (j=-1)
            y0 = optimized.initials[:]
            if self.model.problem.simulate(optimized.params, y0) is None:
                for k, _ in enumerate(self.model.observables):
                    for l, _ in enumerate(self.model.problem.conditions):
                        signaling_metric[i, -1, k, l] = self.quantification[metric](
                            self.model.problem.simulations[k, :, l]
                        )
        sensitivity_coefficients = dlnyi_dlnxj(
            signaling_metric,
            n_file,
            nonzero_indices,
            self.model.observables,
            self.model.problem.conditions,
            rate,
        )

        return sensitivity_coefficients

    def _load_sc(
        self,
        metric: str,
        nonzero_indices: List[int],
    ) -> np.ndarray:
        """
        Load (or calculate) sensitivity coefficients.
        """
        if not os.path.isfile(self._coefficients(metric)):
            os.makedirs(
                os.path.join(
                    self.model.path,
                    "sensitivity_coefficients",
                    "initial_condition",
                ),
                exist_ok=True
                )
            sensitivity_coefficients = self._calc_sensitivity_coefficients(metric, nonzero_indices)
            np.save(
                self._coefficients(metric), sensitivity_coefficients
            )
        else:
            sensitivity_coefficients = np.load(self._coefficients(metric))
            if len(nonzero_indices) != sensitivity_coefficients.shape[1]:
                # User changed options['excluded_initials'] after the last trial
                sensitivity_coefficients = self._calc_sensitivity_coefficients(
                    metric, nonzero_indices
                )
                np.save(self._coefficients(metric), sensitivity_coefficients)

        return sensitivity_coefficients

    def _barplot_sensitivity(
        self,
        metric: str,
        sensitivity_coefficients: np.ndarray,
        nonzero_indices: List[int],
    ) -> None:
        """
        Visualize sensitivity coefficients using barplot.
        """
        os.makedirs(
            os.path.join(self._path_to_figs(metric), "barplot"), exist_ok=True
        )
        # rcParams
        self.model.viz.set_sensitivity_rcParams()

        if len(self._plotting.cmap) < len(self.model.problem.conditions):
            raise ValueError(
                "len(sensitivity_options['cmap']) must be equal to "
                "or greater than len(problem.conditions)."
            )
        for k, obs_name in enumerate(self.model.observables):
            plt.figure(figsize=self._plotting.figsize)
            plt.hlines([0], -self._plotting.width, len(nonzero_indices), "k", lw=1)
            for l, condition in enumerate(self.model.problem.conditions):
                sensitivity_matrix = sensitivity_coefficients[:, :, k, l]
                nan_idx = []
                for m in range(sensitivity_matrix.shape[0]):
                    if np.isnan(sensitivity_matrix[m, :]).any():
                        nan_idx.append(m)
                sensitivity_matrix = np.delete(sensitivity_matrix, nan_idx, axis=0)
                if sensitivity_matrix.size != 0:
                    average = np.mean(sensitivity_matrix, axis=0)
                    if sensitivity_matrix.shape[0] == 1:
                        stdev = np.zeros(sensitivity_matrix.shape[1])
                    else:
                        stdev = np.std(sensitivity_matrix, axis=0, ddof=1)
                    plt.bar(
                        np.arange(len(nonzero_indices)) + l * self._plotting.width,
                        average,
                        yerr=stdev,
                        ecolor=self._plotting.cmap[l],
                        capsize=2,
                        width=self._plotting.width,
                        color=self._plotting.cmap[l],
                        align="center",
                        label=condition,
                    )
            plt.xticks(
                np.arange(len(nonzero_indices))
                + self._plotting.width * 0.5 * (len(self.model.problem.conditions) - 1),
                [
                    self.model.viz.convert_species_name(self.model.species[i])
                    for i in nonzero_indices
                ],
                rotation=90,
            )
            plt.ylabel(
                "Control coefficients on\n" + metric + " (" + obs_name.replace("_", " ") + ")"
            )
            plt.xlim(-self._plotting.width, len(nonzero_indices))
            if self._plotting.legend_kws is not None:
                plt.legend(**self._plotting.legend_kws)
            plt.savefig(
                os.path.join(self._path_to_figs(metric), "barplot", f"{obs_name}")
            )
            plt.close()

    @staticmethod
    def _remove_nan(sensitivity_matrix: np.ndarray, normalize: bool) -> np.ndarray:
        """
        Remove NaN from sensitivity matrix.
        """
        nan_idx = []
        for i in range(sensitivity_matrix.shape[0]):
            if np.isnan(sensitivity_matrix[i, :]).any():
                nan_idx.append(i)
            else:
                pass
            if np.nanmax(np.abs(sensitivity_matrix[i, :])) == 0.0:
                sensitivity_matrix[i, :] = np.zeros(sensitivity_matrix.shape[1])
            else:
                sensitivity_matrix[i, :] = sensitivity_matrix[i, :] / (
                    np.nanmax(np.abs(sensitivity_matrix[i, :])) if normalize else 1
                )

        return np.delete(sensitivity_matrix, nan_idx, axis=0)

    def _heatmap_sensitivity(
        self,
        metric: str,
        sensitivity_coefficients: np.ndarray,
        nonzero_indices: List[int],
    ) -> None:
        """
        Visualize sensitivity coefficients using heatmap.
        """
        os.makedirs(
            os.path.join(self._path_to_figs(metric), "heatmap"), exist_ok=True
        )
        # rcParams
        self.model.viz.set_sensitivity_rcParams()

        for k, obs_name in enumerate(self.model.observables):
            for l, condition in enumerate(self.model.problem.conditions):
                sensitivity_matrix = self._remove_nan(
                    sensitivity_coefficients[:, :, k, l], normalize=False
                )
                if sensitivity_matrix.shape[0] > 1 and not np.all(sensitivity_matrix == 0.0):
                    g = sns.clustermap(
                        data=sensitivity_matrix,
                        center=0,
                        robust=True,
                        method="ward",
                        cmap="RdBu_r",
                        linewidth=0.5,
                        col_cluster=False,
                        figsize=self._plotting.figsize,
                        xticklabels=[
                            self.model.viz.convert_species_name(self.model.species[i])
                            for i in nonzero_indices
                        ],
                        yticklabels=[],
                        # cbar_kws={"ticks": [-1, 0, 1]}
                    )
                    cbar = g.ax_heatmap.collections[0].colorbar
                    cbar.ax.tick_params(labelsize=8)
                    plt.setp(g.ax_heatmap.get_xticklabels(), rotation=90)
                    plt.savefig(
                        os.path.join(
                            self._path_to_figs(metric),
                            "heatmap",
                            f"{condition}_{obs_name}",
                        ),
                        bbox_inches="tight",
                    )
                    plt.close()

    def analyze(self, *, metric: str, style: str, options: dict) -> None:
        """
        Perform sensitivity analysis.
        """
        if options["overwrite"] and os.path.isfile(self._coefficients(metric)):
            os.remove(self._coefficients(metric))
        nonzero_indices = self._get_nonzero_indices(options["excluded_initials"])
        sensitivity_coefficients = self._load_sc(metric, nonzero_indices)
        if style == "barplot":
            self._barplot_sensitivity(
                metric,
                sensitivity_coefficients,
                nonzero_indices,
            )
        elif style == "heatmap":
            if len(nonzero_indices) < 2:
                pass
            else:
                self._heatmap_sensitivity(
                    metric,
                    sensitivity_coefficients,
                    nonzero_indices,
                )
        else:
            raise ValueError("Available styles are: 'barplot', 'heatmap'")



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "uptake",
    "TNF",
    "trigger_iIkk",
    "deact_TNFR",
    "deact_ppIkk",
    "deact_pnNfk",
    "act_Ikk_by_TNF",
    "act_pIkk",
    "act_Ikb_by_Ikk",
    "act_Nfk_by_Ikk",
    "act_Nfk_by_Ikk_complex",
    "act_Ikb_complex",
    "form_complex",
    "form_complex_nuc",
    "ext_nNfkIkb",
    "Vnuc",
    "split_NfkpIkb",
    "split_NfkIkb",
    "int_Nfk",
    "int_Ikb",
    "eta_int_pNfk",
    "degrad_Ikb",
    "degrad_mIkb",
    "degrad_RnaA20",
    "degrad_A20",
    "prod_Ikb",
    "prod_mIkb_by_nNfk",
    "build_RnaA20",
    "build_A20",
    "shuttle_RnaA20",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



import os
import time
import warnings
from math import isnan

import numpy as np

from ..exec_model import ExecModel, ModelObject
from .step import OptimizeStep


class OptimizeWarning(UserWarning):
    pass


def _check_unknown_options(unknown_options: dict) -> None:
    if unknown_options:
        msg = ", ".join(map(str, unknown_options.keys()))
        warnings.warn(f"Unknown optimization options: {msg}", OptimizeWarning)


class GeneticAlgorithmInit(ExecModel):
    """
    Use genetic algorithm to estimate model parameters from experimental measurements.
    """

    def __init__(
        self,
        model: ModelObject,
        popsize: int,
        max_generation: int,
        initial_threshold: float,
        allowable_error: float,
        local_search_method: str,
        n_children: int,
        maxiter: int,
        workers: int,
        overwrite: bool,
        **unknown_options,
    ) -> None:
        super().__init__(model)
        self.search_rgn: np.ndarray = self.model.problem.get_region()
        self.n_population: int = int(popsize * self.search_rgn.shape[1])
        self.n_gene: int = self.search_rgn.shape[1]
        self.max_generation: int = max_generation
        self.initial_threshold: float = initial_threshold
        self.allowable_error: float = allowable_error
        self.local_search_method: str = local_search_method.lower()
        self.n_children: int = n_children
        self.maxiter: int = maxiter
        self.workers: int = workers
        self.overwrite: bool = overwrite
        
        _check_unknown_options(unknown_options)
        if self.n_population < self.n_gene + 2:
            raise ValueError(f"n_population must be larger than {self.n_gene + 2:d}")

    def run(self, nth_paramset: int) -> None:
        """
        Run GA.
        """
        os.makedirs(
            os.path.join(
                self.model.path,
                "out",
            ),
            exist_ok=True,
        )
        if not self.overwrite and os.path.isdir(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
            )
        ):
            raise FileExistsError(
                "Set options['overwrite']=True to overwrite "
                + os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                )
            )
        else:
            try:
                files = os.listdir(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                    )
                )
                for file in files:
                    if any(map(file.__contains__, (".npy", ".log"))):
                        os.remove(
                            os.path.join(
                                self.model.path,
                                "out",
                                f"{nth_paramset:d}",
                                f"{file}",
                            )
                        )
            except FileNotFoundError:
                os.mkdir(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                    )
                )
        np.random.seed(time.time_ns() * nth_paramset % 2 ** 32)
        warnings.filterwarnings("ignore")
        self._ga_v2(nth_paramset)

    def _set_initial(self, nth_paramset: int) -> np.ndarray:
        """
        Generate initial populations.
        """
        population = np.full((self.n_population, self.n_gene + 1), np.inf)
        with open(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "optimization.log",
            ),
            mode="w",
        ) as f:
            f.write("Generating the initial population. . .\n")
        for i in range(self.n_population):
            while self.initial_threshold <= population[i, -1] or isnan(population[i, -1]):
                population[i, : self.n_gene] = np.random.rand(self.n_gene)
                population[i, -1] = self.get_obj_val(population[i, : self.n_gene])
            with open(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "optimization.log",
                ),
                mode="a",
            ) as f:
                f.write(f"{i + 1:d} / {self.n_population:d}\n")
        population = population[np.argsort(population[:, -1]), :]

        return population

    def _ga_v2(self, nth_paramset: int) -> None:
        """
        Minimize the distance between model simulation and experimental data
        using genetic algorithm.

        Notes
        -----
        ga_v2 optimizes an objective function through the following procedure.

        1. Initialization
            As an initial population, create np individuals randomly.
            `ga_v2` also represents individuals as n-dimensional real number
            vectors, where n is the dimension of the search space. Set
            Generation to 0, and set the iteration number of converging
            operations `Niter` to 1.

        2. Selection for reproduction
            As parents for the recombination operator, ENDX, select `m`
            individuals, `p1`, `p2`, . . . , `pm`, without replacement from the
            population.

        3. Generation of offsprings
            Generate `Nc` children by applying ENDX to the selected parents.
            This algorithm assigns the worst objective value to the children.

        4. Local Search
            Apply the local search method to the best individual in a family
            consisting of the two parents, i.e., `p1` and `p2`, and their children.
            Note here that the children are assumed to have the worst objective
            value. Thus, whenever the objective values of the two parents have
            been actually computed in previous generations, the algorithm
            applies the local search to either of the parents. When all of the
            individuals in the family have the same objective value, on the
            other hand, the local search is applied to a randomly selected
            individual from the family.

        5. Selection for survival
            Select two individuals from the family. The first selected
            individual should be the individual with the best objective value,
            and the second should be selected randomly. Then, replace the two
            parents (`p1` and `p2`) with the selected individuals. Note that the
            individual to which the local search has been applied in the
            previous step is always selected as the best.

        6. Application of ENDX/MGG
            To achieve a good search performance, `ga_v2` optimizes a function,
            gradually narrowing the search space. For this purpose, the
            converging phase slightly converges the population by repeating the
            following procedure Niter times.

            1.
                Select m individuals without replacement from the population.
                The selected individuals, expressed here as `p1`, `p2`, . . . , `pm`,
                are used as the parents for an extended normal distribution
                crossover (ENDX) applied in the next step.

            2.
                Generate `Nc` children by applying ENDX to the parents selected
                in the previous step. To reduce the computational cost, `ga_v2`
                forgoes any computation of the objective values of the `Nc`
                individuals generated here. Instead, the algorithm assigns the
                newly generated children a single objective value, one which is
                inferior to the objective values of any of the possible
                candidate solutions.

            3.
                Select two individuals from a family containing the two
                parents, i.e., `p1` and `p2`, and their children. The first
                selected individual should be the one with the best objective
                value, and the second should be selected randomly. Then,
                replace the two parents with the selected individuals.

        7. Adaptation of Niter
            If the best individual has not improved during the last np
            generations, `Niter` <- 2 * `Niter`. Otherwise, set `Niter` to 1.

        8. Termination
            Stop if the halting criteria are satisfied.
            Otherwise, `Generation` <- `Generation` + 1, and return to the step 2.
        """
        step = OptimizeStep(
            self.get_obj_val,
            self.n_population,
            self.n_gene,
            self.n_children,
            self.maxiter,
            self.workers,
        )
        n_iter = 1
        n0 = np.empty(3 * self.n_population)

        population = self._set_initial(nth_paramset)
        n0[0] = population[0, -1]

        with open(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "optimization.log",
            ),
            mode="a",
        ) as f:
            f.write(
                "\n----------------------------------------\n\n"
                + f"Generation1: Best Fitness = {population[0, -1]:e}\n"
            )
        best_individual = self.model.problem.gene2val(population[0, : self.n_gene])
        best_fitness = population[0, -1]

        np.save(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "generation.npy",
            ),
            1,
        )
        np.save(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "fit_param1.npy",
            ),
            best_individual,
        )
        np.save(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "best_fitness.npy",
            ),
            best_fitness,
        )
        if population[0, -1] <= self.allowable_error:
            return

        generation = 1
        while generation < self.max_generation:
            ip = np.random.choice(self.n_population, self.n_gene + 2, replace=False)
            population = step.converging(ip, population)
            population = step.local_search(ip, population, self.local_search_method)
            for _ in range(n_iter - 1):
                ip = np.random.choice(self.n_population, self.n_gene + 2, replace=False)
                population = step.converging(ip, population)
            # Adaptation of n_iter
            if generation % len(n0) == len(n0) - 1:
                n0[-1] = population[0, -1]
                if n0[0] == n0[-1]:
                    n_iter *= 2
                else:
                    n_iter = 1
            else:
                n0[generation % len(n0)] = population[0, -1]

            best_individual = self.model.problem.gene2val(population[0, : self.n_gene])
            if population[0, -1] < best_fitness:
                np.save(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        "generation.npy",
                    ),
                    generation + 1,
                )
                np.save(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        f"fit_param{generation + 1:d}.npy",
                    ),
                    best_individual,
                )
            best_fitness = population[0, -1]
            np.save(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "best_fitness.npy",
                ),
                best_fitness,
            )
            np.save(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "count_num.npy",
                ),
                generation + 1,
            )
            with open(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "optimization.log",
                ),
                mode="a",
            ) as f:
                f.write(f"Generation{generation + 1:d}: " f"Best Fitness = {best_fitness:e}\n")
            if population[0, -1] <= self.allowable_error:
                break

            generation += 1

        return


class GeneticAlgorithmContinue(ExecModel):
    """
    Continue optimization from where you stopped in the last parameter estimation.
    """

    def __init__(
        self,
        model: ModelObject,
        popsize: int,
        max_generation: int,
        initial_threshold: float,
        allowable_error: float,
        local_search_method: str,
        n_children: int,
        maxiter: int,
        workers: int,
        p0_bounds: list,
        **unknown_options,
    ) -> None:
        super().__init__(model)
        self.search_rgn: np.ndarray = self.model.problem.get_region()
        self.n_population: int = int(popsize * self.search_rgn.shape[1])
        self.n_gene: int = self.search_rgn.shape[1]
        self.max_generation: int = max_generation
        self.initial_threshold: float = initial_threshold
        self.allowable_error: float = allowable_error
        self.local_search_method: str = local_search_method.lower()
        self.n_children: int = n_children
        self.maxiter: int = maxiter
        self.workers: int = workers
        self.p0_bounds: list = p0_bounds

        _check_unknown_options(unknown_options)
        if self.n_population < self.n_gene + 2:
            raise ValueError(f"n_population must be larger than {self.n_gene + 2:d}")

    def run(self, nth_paramset: int) -> None:
        """
        Run GA.
        """
        np.random.seed(time.time_ns() * nth_paramset % 2 ** 32)
        warnings.filterwarnings("ignore")
        self._ga_v2_continue(nth_paramset)

    def _set_continue(self, nth_paramset: int) -> np.ndarray:
        """
        Generate initial populations.
        """
        best_generation = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "generation.npy",
            )
        )
        best_individual = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                f"fit_param{int(best_generation):d}.npy",
            )
        )
        population = np.full((self.n_population, self.n_gene + 1), np.inf)

        with open(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "optimization.log",
            ),
            mode="a",
        ) as f:
            f.write("\n#####\n##### Continue\n#####\nGenerating the initial population. . .\n")
        for i in range(self.n_population):
            while self.initial_threshold <= population[i, -1]:
                population[i, : self.n_gene] = self._encode_bestIndivVal2randGene(best_individual)
                population[i, : self.n_gene] = np.clip(population[i, : self.n_gene], 0.0, 1.0)
                population[i, -1] = self.get_obj_val(population[i, : self.n_gene])
            with open(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "optimization.log",
                ),
                mode="a",
            ) as f:
                f.write(f"{i + 1:d} / {self.n_population:d}\n")
        population = population[np.argsort(population[:, -1]), :]

        return population

    def _encode_bestIndivVal2randGene(self, best_individual: np.ndarray) -> np.ndarray:
        rand_gene = (
            np.log10(
                best_individual
                * 10
                ** (
                    np.random.rand(len(best_individual))
                    * np.log10(self.p0_bounds[1] / self.p0_bounds[0])
                    + np.log10(self.p0_bounds[0])
                )
            )
            - self.search_rgn[0, :]
        ) / (self.search_rgn[1, :] - self.search_rgn[0, :])

        return rand_gene

    def _ga_v2_continue(self, nth_paramset: int) -> None:
        step = OptimizeStep(
            self.get_obj_val,
            self.n_population,
            self.n_gene,
            self.n_children,
            self.maxiter,
            self.workers,
        )
        n_iter = 1
        n0 = np.empty(3 * self.n_population)

        count_num = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "count_num.npy",
            )
        )
        best_generation = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "generation.npy",
            )
        )
        best_individual = np.load(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                f"fit_param{int(best_generation):d}.npy",
            )
        )
        best_fitness = self.model.problem.objective(best_individual)

        if self.max_generation <= count_num:
            raise ValueError(f"max_generation should be larger than {int(count_num):d}")

        population = self._set_continue(nth_paramset)
        if best_fitness < population[0, -1]:
            population[0, : self.n_gene] = self.model.problem.val2gene(best_individual)
            population[0, -1] = best_fitness
        else:
            best_individual = self.model.problem.gene2val(population[0, : self.n_gene])
            best_fitness = population[0, -1]
            np.save(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    f"fit_param{int(count_num) + 1:d}.npy",
                ),
                best_individual,
            )
        with open(
            os.path.join(
                self.model.path,
                "out",
                f"{nth_paramset:d}",
                "optimization.log",
            ),
            mode="a",
        ) as f:
            f.write(
                "\n----------------------------------------\n\n"
                f"Generation{int(count_num) + 1:d}: "
                f"Best Fitness = {best_fitness:e}\n"
            )
        n0[0] = population[0, -1]

        if population[0, -1] <= self.allowable_error:
            return

        generation = 1 + int(count_num)
        while generation < self.max_generation:
            ip = np.random.choice(self.n_population, self.n_gene + 2, replace=False)
            population = step.converging(ip, population)
            population = step.local_search(ip, population, self.local_search_method)
            for _ in range(n_iter - 1):
                ip = np.random.choice(self.n_population, self.n_gene + 2, replace=False)
                population = step.converging(ip, population)
            if generation % len(n0) == len(n0) - 1:
                n0[-1] = population[0, -1]
                if n0[0] == n0[-1]:
                    n_iter *= 2
                else:
                    n_iter = 1
            else:
                n0[generation % len(n0)] = population[0, -1]

            best_individual = self.model.problem.gene2val(population[0, : self.n_gene])
            if population[0, -1] < best_fitness:
                np.save(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        "generation.npy",
                    ),
                    generation + 1,
                )
                np.save(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        f"fit_param{generation + 1:d}.npy",
                    ),
                    best_individual,
                )
                np.save(
                    os.path.join(
                        self.model.path,
                        "out",
                        f"{nth_paramset:d}",
                        "best_fitness",
                    ),
                    best_fitness,
                )
            best_fitness = population[0, -1]
            np.save(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "count_num.npy",
                ),
                generation + 1,
            )
            with open(
                os.path.join(
                    self.model.path,
                    "out",
                    f"{nth_paramset:d}",
                    "optimization.log",
                ),
                mode="a",
            ) as f:
                f.write(f"Generation{generation + 1:d}: " f"Best Fitness = {best_fitness:e}\n")
            if population[0, -1] <= self.allowable_error:
                break

            generation += 1

        return



"""
Figure properties for visualizing time-course simulations and sensitivity analysis results.

Example code: Nakakuki_Cell_2010_

.. _Nakakuki_Cell_2010: https://github.com/biomass-dev/biomass/blob/master/biomass/models/Nakakuki_Cell_2010/viz.py
"""

from typing import List, Optional, Union

from matplotlib.colors import ListedColormap
from matplotlib.lines import Line2D

__all__ = ["SingleObservable", "MultipleObservables", "SensitivityOptions"]


class SingleObservable(object):
    """
    Visualization options for time-course simulation (single-observable).

    Attributes
    ----------
    divided_by : int or float (default: 1)
        Convert time unit. (e.g. sec -> min).

    figsize : tuple (default: (4, 3))
        Width, height in inches.

    xlim : tuple
        Set the x limits of the current axes.

    xticks : list (default: None)
        Set the current tick locations of the x-axis.

    xlabel : str (default: 'Time')
        Set the label for the x-axis.

    ylim : tuple
        Set the y limits of the current axes.

    yticks : list (default: None)
        Set the current tick locations of the y-axis.

    ylabel : str (default: obs_name.replace('_', ' '))
        Set the label for the y-axis.

    exp_data : bool (default: True)
        if False, experimental data will not be shown.

    legend_kws : dict, optional
        Keyword arguments to pass to ``matplotlib.pyplot.legend()``.

    cmap : list or tuple
        Set colormap.

    shape : list or tuple of strings (default: `Line2D.filled_markers`)
        Set markers.

    dont_show : list of strings
        Set conditions you don't want to plot.
    """
    
    def __init__(self, cm: ListedColormap, obs_name: str) -> None:
        self.divided_by: int = 1
        self.figsize: tuple = (4, 3)
        self.xlim: tuple = ()
        self.xticks: Optional[list] = None
        self.xlabel: str =  "Time"
        self.ylim: tuple = ()
        self.yticks: Optional[list] = None
        self.ylabel: str =  obs_name.replace("__", "\n").replace("_", " ")
        self.exp_data: bool = True
        self.legend_kws: Optional[dict] = None
        self.cmap: list =  [cm.colors[j] for j in range(10)]
        self.shape: Union[list, tuple] = Line2D.filled_markers
        self.dont_show: List[str] = []


class MultipleObservables(object):
    """
    Visualization options for time-course simulation (multi-observables).

    Attributes
    ----------
    fname : str
        Output file name.
    
    figsize : tuple (default: (4, 3))
        Figure size.
    
    observables : list of strings
        Specify which observables to plot.
    
    condition : str, optional
        Specify which simulation condition to be plotted.
    
    xlim : tuple
        Set the x limits of the current axes.

    xticks : list (default: None)
        Set the current tick locations of the x-axis.

    xlabel : str (default: 'Time')
        Set the label for the x-axis.

    ylim : tuple
        Set the y limits of the current axes.

    yticks : list (default: None)
        Set the current tick locations of the y-axis.

    ylabel : str (default: '')
        Set the label for the y-axis.
    
    legend_kws : dict, optional
        Keyword arguments to pass to ``matplotlib.pyplot.legend()``.

    cmap : list or tuple
        Set colormap.

    shape : list or tuple of strings (default: `Line2D.filled_markers`)
        Set markers.
    """

    def __init__(self, cm: ListedColormap) -> None:
        self.fname: str =  "multi_observables"
        self.figsize: tuple = (4, 3)
        self.observables: List[str] = []
        self.condition: Optional[str] = None
        self.xlim: tuple = ()
        self.xticks: Optional[list] = None
        self.xlabel: str =  "Time"
        self.ylim: tuple = ()
        self.yticks: Optional[list] = None
        self.ylabel: str = ""
        self.legend_kws: Optional[dict] = dict(
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
            borderaxespad=0,
            labelspacing=1.25,
            frameon=False,
            fontsize=12,
        )
        self.cmap: list =  [cm.colors[j] for j in range(10)]
        self.shape: Union[list, tuple] = Line2D.filled_markers


class SensitivityOptions(object):
    """
    Visualization options for sensitivity analysis results.

    Attributes
    ----------
    figsize : tuple
        Figure size.

    width : float
        The width(s) of the bars when choosing `style` == 'barplot'.

    legend_kws : dict, optional
        Keyword arguments to pass to ``matplotlib.pyplot.legend()``.
        
    cmap : list or tuple
        Set colormap.
    """
    def __init__(self, cm: ListedColormap) -> None:
        self.figsize: tuple = (12, 5)
        self.width: float = 0.3
        self.legend_kws: Optional[dict] = dict(loc="upper left", frameon=False)
        self.cmap: list =  [cm.colors[j] for j in range(10)]



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation


class Observable(DifferentialEquation):
    """
    Correlating model simulations and experimental measurements.

    Attributes
    ----------
    obs_names : list of strings
        Names of model observables.

    t : range
        Simulation time span.

    conditions : list of strings
        Expetimental conditions.

    simulations : numpy.ndarray
        The numpy array to store simulation results.

    normalization : nested dict
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        super(Observable, self).__init__(perturbation={})
        self.obs_names: list = [
            "pAKT",
            "pS6K",
            "pGSK3B",
            "G6Pase",
        ]
        self.t: range = range(480 + 1)  # min
        self.conditions: list = [
            "Insulin 0.01",
            "Insulin 0.03",
            "Insulin 0.1",
            "Insulin 0.3",
            "Insulin 1.0",
        ]
        self.simulations: np.ndarray = np.empty(
            (len(self.obs_names), len(self.t), len(self.conditions))
        )
        self.normalization: dict = {}
        self.experiments: list = [None] * len(self.obs_names)
        self.error_bars: list = [None] * len(self.obs_names)

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation

        y0[V.Ins] = 0.01  # 0.01 nM of insulin during starvation
        sol = solve_ode(self.diffeq, y0, range(2400 + 1), tuple(x))
        if sol is None:
            return False
        else:
            y0 = sol.y[:, -1].tolist()

        for i, condition in enumerate(self.conditions):
            if condition == "Insulin 0.01":
                y0[V.Ins] = 0.01
            elif condition == "Insulin 0.03":
                y0[V.Ins] = 0.03
            elif condition == "Insulin 0.1":
                y0[V.Ins] = 0.1
            elif condition == "Insulin 0.3":
                y0[V.Ins] = 0.3
            elif condition == "Insulin 1.0":
                y0[V.Ins] = 1.0

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[self.obs_names.index("pAKT"), :, i] = sol.y[V.pAKT, :]
                self.simulations[self.obs_names.index("pS6K"), :, i] = (
                    sol.y[V.pS6K, :] * 83.8672192461257
                )
                self.simulations[self.obs_names.index("pGSK3B"), :, i] = (
                    sol.y[V.pGSK3B, :] * 0.111097316860158
                )
                self.simulations[self.obs_names.index("G6Pase"), :, i] = (
                    sol.y[V.G6Pase, :] * 0.0363622452066626
                )

    def set_data(self):
        pass

    def get_timepoint(self, obs_name):
        if obs_name in self.obs_names:
            return []



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation


class Observable(DifferentialEquation):
    """
    Correlating model simulations and experimental measurements.

    Attributes
    ----------
    obs_names : list of strings
        Names of model observables.

    t : range
        Simulation time span.

    conditions : list of strings
        Expetimental conditions.

    simulations : numpy.ndarray
        The numpy array to store simulation results.

    normalization : nested dict
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        super(Observable, self).__init__(perturbation={})
        self.obs_names: list = [
            "Per_mRNA",
            "Cry_mRNA",
            "Bmal1_mRNA",
        ]
        self.t: range = range(72 + 1)
        self.conditions: list = ["DD"]
        self.simulations: np.ndarray = np.empty(
            (len(self.obs_names), len(self.t), len(self.conditions))
        )
        self.normalization: dict = {}
        self.experiments: list = [None] * len(self.obs_names)
        self.error_bars: list = [None] * len(self.obs_names)

    def simulate(self, x, y0, _perturbation={}):
        if _perturbation:
            self.perturbation = _perturbation
        for i, condition in enumerate(self.conditions):
            if condition == "DD":
                pass

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[self.obs_names.index("Per_mRNA"), :, i] = sol.y[V.MP, :]
                self.simulations[self.obs_names.index("Cry_mRNA"), :, i] = sol.y[V.MC, :]
                self.simulations[self.obs_names.index("Bmal1_mRNA"), :, i] = sol.y[V.MB, :]

    def set_data(self):
        pass

    def get_timepoint(self, obs_name):
        if obs_name in self.obs_names:
            return []



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "TNFR",
    "Ikk",
    "pIkk",
    "ppIkk",
    "iIkk",
    "NfkIkb",
    "NfkpIkb",
    "pNfkIkb",
    "pNfkpIkb",
    "pNfk",
    "Nfk",
    "pIkb",
    "Ikb",
    "mIkb",
    "nIkb",
    "pnNfk",
    "nNfk",
    "nNfkIkb",
    "RnaA20_1",
    "RnaA20",
    "A20",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .set_model import DifferentialEquation


class Observable(DifferentialEquation):
    """
    Correlating model simulations and experimental measurements.

    Attributes
    ----------
    obs_names : list of strings
        Names of model observables.

    t : range
        Simulation time span.

    conditions : list of strings
        Expetimental conditions.

    simulations : numpy.ndarray
        The numpy array to store simulation results.

    normalization : nested dict
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If None, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in sim.conditions will be used.

    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        super(Observable, self).__init__(perturbation={})
        self.obs_names: list = [
            "biphosphorylated_MAPK",
            "unphosphorylated_MAPK",
        ]
        self.t: range = range(150 * 60 + 1)
        self.conditions: list = ["control"]
        self.simulations: np.ndarray = np.empty(
            (len(self.obs_names), len(self.t), len(self.conditions))
        )
        self.normalization: dict = {}
        self.experiments: list = [None] * len(self.obs_names)
        self.error_bars: list = [None] * len(self.obs_names)

    def simulate(self, x, y0, _perturbation=None):
        if _perturbation is not None:
            self.perturbation = _perturbation
        for i, condition in enumerate(self.conditions):
            if condition == "control":
                pass
            """
            elif condition == 'cooperative':
                x[C.n] = 2
                x[C.KI] = 18
                x[C.K1] = 50
                x[C.K2] = 40
                x[C.K3] = 100
                x[C.K4] = 100
                x[C.K5] = 100
                x[C.K6] = 100
                x[C.K7] = 100
                x[C.K8] = 100
                x[C.K9] = 100
                x[C.K10] = 100
                x[C.V9] = 1.25
                x[C.V10] = 1.25
            """

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[self.obs_names.index("biphosphorylated_MAPK"), :, i] = sol.y[
                    V.MAPK_PP, :
                ]
                self.simulations[self.obs_names.index("unphosphorylated_MAPK"), :, i] = sol.y[
                    V.MAPK, :
                ]

    def set_data(self):
        # Test data
        self.experiments[self.obs_names.index("biphosphorylated_MAPK")] = {
            "control": [
                10.0,
                298.59241599,
                298.57164292,
                295.16057672,
                211.52330623,
                80.97440738,
                84.52996829,
                296.19437698,
                254.48224797,
                122.22861484,
                36.96467096,
                296.55589221,
                284.81739793,
                166.22074898,
                47.45788287,
                215.79175866,
                293.95196966,
                210.80086084,
                80.87144427,
                84.78789794,
                296.18270875,
                254.19844509,
                121.93215568,
                37.04378877,
                296.56176029,
                284.69692824,
                165.91228275,
                47.27840604,
                216.91315533,
                293.92385296,
            ]
        }
        self.experiments[self.obs_names.index("unphosphorylated_MAPK")] = {
            "control": [
                2.80000000e02,
                1.02084698e-01,
                1.04142327e-01,
                7.54905842e-01,
                5.03960433e01,
                1.60518127e02,
                1.45421258e02,
                5.24026172e-01,
                1.98605774e01,
                1.23225761e02,
                2.02633306e02,
                4.51355981e-01,
                3.98736964e00,
                8.59280677e01,
                1.92406000e02,
                1.72867943e01,
                1.05444189e00,
                5.06418164e01,
                1.60322956e02,
                1.44990726e02,
                5.26516648e-01,
                2.00396352e01,
                1.23483781e02,
                2.02532132e02,
                4.50047126e-01,
                4.03300324e00,
                8.61819825e01,
                1.92586475e02,
                1.66298172e01,
                1.06168913e00,
            ]
        }

    def get_timepoint(self, obs_name):
        if obs_name in self.obs_names:
            return [60 * i for i in range(0, 150, 5)]



from matplotlib import pyplot as plt

from biomass.plotting import *

from .observable import Observable


class Visualization(Observable):
    """
    Plotting parameters for customizing figure properties.

    Attributes
    ----------
    cm : matplotlib.colors.ListedColormap (default: `plt.cm.get_cmap('tab10')`)
        Choosing colormaps for `cmap`.
    single_observable_options : list of SingleObservable
        Visualization options for time-course simulation (single-observable).
    multiple_observables_options : MultipleObservables
        Visualization options for time-course simulation (multi-observables).
    sensitivity_options : SensitivityOptions
        Visualization options for sensitivity analysis results.
    """

    def __init__(self):
        super().__init__()

        self.cm = plt.cm.get_cmap("tab10")
        self.single_observable_options = [
            SingleObservable(self.cm, obs_name) for obs_name in self.obs_names
        ]
        self.multiple_observables_options = MultipleObservables(self.cm)
        self.sensitivity_options = SensitivityOptions(self.cm)

    def get_single_observable_options(self):

        for i, _ in enumerate(self.obs_names):
            self.single_observable_options[i].xticks = [200 * i for i in range(4)]
            self.single_observable_options[i].xlabel = "time (min)"

        return self.single_observable_options

    def get_multiple_observables_options(self):

        return self.multiple_observables_options
    
    def get_sensitivity_options(self):

        return self.sensitivity_options

    @staticmethod
    def set_timecourse_rcParams():
        """figure/simulation"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.5
        plt.rcParams["xtick.major.width"] = 1.5
        plt.rcParams["ytick.major.width"] = 1.5
        plt.rcParams["lines.linewidth"] = 1.8
        plt.rcParams["lines.markersize"] = 12
        plt.rcParams["savefig.bbox"] = "tight"
        # plt.rcParams["savefig.format"] = "pdf"
        # plt.rcParams['font.family'] = 'Arial'
        # plt.rcParams['mathtext.fontset'] = 'custom'
        # plt.rcParams['mathtext.it'] = 'Arial:italic'

    @staticmethod
    def set_sensitivity_rcParams():
        """figure/sensitivity"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.2
        plt.rcParams["xtick.major.width"] = 1.2
        plt.rcParams["ytick.major.width"] = 1.2
        plt.rcParams["savefig.bbox"] = "tight"
        # plt.rcParams["savefig.format"] = "pdf"
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def convert_species_name(name):
        """figure/sensitivity/initial_condition
        - Sensitivity for species with nonzero initial conditions
        """
        return name



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "V1",
    "n",
    "KI",
    "K1",
    "V2",
    "K2",
    "k3",
    "K3",
    "k4",
    "K4",
    "V5",
    "K5",
    "V6",
    "K6",
    "k7",
    "K7",
    "k8",
    "K8",
    "V9",
    "K9",
    "V10",
    "K10",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



import numpy as np

from biomass.estimation import convert_scale, initialize_search_param

from .name2idx import C, V
from .set_model import initial_values, param_values


class SearchParam(object):
    """Specify model parameters and/or initial values to optimize"""

    def __init__(self):
        # parameters
        self.idx_params: list = [
            C.V1,
            # C.n,
            C.KI,
            C.K1,
            C.V2,
            C.K2,
            C.k3,
            C.K3,
            C.k4,
            C.K4,
            C.V5,
            C.K5,
            C.V6,
            C.K6,
            C.k7,
            C.K7,
            C.k8,
            C.K8,
            C.V9,
            C.K9,
            C.V10,
            C.K10,
        ]

        # initial values
        self.idx_initials: list = [
            # V.(specie)
        ]

    def get_region(self):
        x = param_values()
        y0 = initial_values()

        search_param = initialize_search_param(
            parameters=C.NAMES,
            species=V.NAMES,
            param_values=x,
            initial_values=y0,
            estimated_params=self.idx_params,
            estimated_initials=self.idx_initials,
        )

        search_rgn = np.zeros((2, len(x) + len(y0)))
        # Default: 0.1 ~ 10
        for i, j in enumerate(self.idx_params):
            search_rgn[0, j] = search_param[i] * 0.5  # lower bound
            search_rgn[1, j] = search_param[i] * 2.0  # upper bound
        # Default: 0.5 ~ 2
        for i, j in enumerate(self.idx_initials):
            search_rgn[0, j + len(x)] = search_param[i + len(self.idx_params)] * 0.5  # lower bound
            search_rgn[1, j + len(x)] = search_param[i + len(self.idx_params)] * 2.0  # upper bound

        # search_rgn[:,C.parameter] = [lower_bound, upper_bound]
        # search_rgn[:,V.specie+len(x)] = [lower_bound, upper_bound]

        search_rgn = convert_scale(
            region=search_rgn,
            parameters=C.NAMES,
            species=V.NAMES,
            estimated_params=self.idx_params,
            estimated_initials=self.idx_initials,
        )

        return search_rgn

    def update(self, indiv):
        x = param_values()
        y0 = initial_values()

        for i, j in enumerate(self.idx_params):
            x[j] = indiv[i]
        for i, j in enumerate(self.idx_initials):
            y0[j] = indiv[i + len(self.idx_params)]

        return x, y0

    def gene2val(self, indiv_gene):
        search_rgn = self.get_region()
        indiv = 10 ** (indiv_gene * (search_rgn[1, :] - search_rgn[0, :]) + search_rgn[0, :])

        return indiv

    def val2gene(self, indiv):
        search_rgn = self.get_region()
        indiv_gene = (np.log10(indiv) - search_rgn[0, :]) / (search_rgn[1, :] - search_rgn[0, :])

        return indiv_gene



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "Rec_act",
    "S2tot",
    "S3tot",
    "S4tot",
    "S_dephos",
    "S_dephosphos",
    "S_phos",
    "Ski_act3",
    "Ski_act2",
    "Ski_act1",
    "Ski_inh3",
    "Ski_inh2",
    "Ski_inh1",
    "Ski_turn",
    "Skil_act3",
    "Skil_act2",
    "Skil_act1",
    "Skil_inh3",
    "Skil_inh2",
    "Skil_inh1",
    "Skil_turn",
    "Dnmt3a_act3",
    "Dnmt3a_act2",
    "Dnmt3a_act1",
    "Dnmt3a_inh3",
    "Dnmt3a_inh2",
    "Dnmt3a_inh1",
    "Dnmt3a_turn",
    "Sox4_act3",
    "Sox4_act2",
    "Sox4_act1",
    "Sox4_inh3",
    "Sox4_inh2",
    "Sox4_inh1",
    "Sox4_turn",
    "Jun_act3",
    "Jun_act2",
    "Jun_act1",
    "Jun_inh3",
    "Jun_inh2",
    "Jun_inh1",
    "Jun_turn",
    "Smad7_act3",
    "Smad7_act2",
    "Smad7_act1",
    "Smad7_inh3",
    "Smad7_inh2",
    "Smad7_inh1",
    "Smad7_turn",
    "Klf10_act3",
    "Klf10_act2",
    "Klf10_act1",
    "Klf10_inh3",
    "Klf10_inh2",
    "Klf10_inh1",
    "Klf10_turn",
    "Bmp4_act3",
    "Bmp4_act2",
    "Bmp4_act1",
    "Bmp4_inh3",
    "Bmp4_inh2",
    "Bmp4_inh1",
    "Bmp4_turn",
    "Cxcl15_act3",
    "Cxcl15_act2",
    "Cxcl15_act1",
    "Cxcl15_inh3",
    "Cxcl15_inh2",
    "Cxcl15_inh1",
    "Cxcl15_turn",
    "Dusp5_act3",
    "Dusp5_act2",
    "Dusp5_act1",
    "Dusp5_inh3",
    "Dusp5_inh2",
    "Dusp5_inh1",
    "Dusp5_turn",
    "Tgfa_act3",
    "Tgfa_act2",
    "Tgfa_act1",
    "Tgfa_inh3",
    "Tgfa_inh2",
    "Tgfa_inh1",
    "Tgfa_turn",
    "Pdk4_act3",
    "Pdk4_act2",
    "Pdk4_act1",
    "Pdk4_inh3",
    "Pdk4_inh2",
    "Pdk4_inh1",
    "Pdk4_turn",
    "init_Rec",
    "k_on_223",
    "k_on_224",
    "k_on_233",
    "k_on_234",
    "k_on_244",
    "k_on_334",
    "k_on_344",
    "k_on_222",
    "k_on_333",
    "k_on_444",
    "kdiss_SS",
    "pRec_degind",
    "sd_Bmp4",
    "sd_Cxcl15",
    "sd_Dnmt3a",
    "sd_Dusp5",
    "sd_Jun",
    "sd_Klf10",
    "sd_Pdk4",
    "sd_Ski",
    "sd_Skil",
    "sd_Smad7",
    "sd_Sox4",
    "sd_Tgfa",
    "gene_turn",
    "gene_act1",
    "gene_act2",
    "gene_act3",
    "gene_inh1",
    "gene_inh2",
    "gene_inh3",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



import os
import shutil
import sys
import warnings
from dataclasses import dataclass
from typing import Callable, List, Union

import numpy as np

from ..exec_model import ModelObject


class _Logger(object):
    """
    Duplicate stdout to _external/optimization.log.
    """

    def __init__(self, model_path: str):
        self.terminal = sys.stdout
        self.log = open(
            os.path.join(model_path, "out", "_external", "optimization.log"),
            mode="w",
            encoding="utf-8",
        )

    def write(self, message: str):
        self.terminal.write(message)
        self.log.write(message)


@dataclass
class ExternalOptimizer(object):
    """
    Use external optimization method for parameterization of a mechanistic model.

    Attributes
    ----------

    model : ModelObject
        The BioMASS model object.

    optimize : Callable
        The external optimizer, e.g., `scipy.optimize.differential_evolution`.
    """

    model: ModelObject
    optimize: Callable

    def __post_init__(self):
        os.makedirs(os.path.join(self.model.path, "out", "_external"), exist_ok=True)
        self.default_stdout = sys.stdout

    def run(self, *args, **kwargs):
        """
        Execute the external optimizer.
        """
        try:
            sys.stdout = _Logger(self.model.path)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                res = self.optimize(*args, **kwargs)
        finally:
            sys.stdout = self.default_stdout
        return res

    def import_solution(self, x: Union[np.ndarray, List[float]], x_id: int = 0) -> None:
        """
        Import the solution of the optimization to the model.
        The solution vector `x` will be saved to `path_to_model`/out/`x_id`/.
        Use `biomass.run_simulation` to visualize the optimization result.

        Parameters
        ----------
        x : Union[np.ndarray, List[float]]
            The solution of the optimization.
        x_id : int (default: 0)
            Index of the parameter set.

        Examples
        --------
        >>> from scipy.optimize import differential_evolution
        >>> from biomass import Model
        >>> from biomass.models import Nakakuki_Cell_2010
        >>> model = Model(Nakakuki_Cell_2010.__package__).create()
        >>> optimizer = ExternalOptimizer(model, differential_evolution)
        >>> res = optimizer.run(
        ...     model.problem.objective,
        ...     model.problem.bounds,
        ...     strategy="best2bin",
        ...     maxiter=100,
        ...     tol=1e-4,
        ...     mutation=0.1,
        ...     recombination=0.5,
        ...     disp=True,
        ...     polish=False,
        ...     workers=-1,
        ... )

        differential_evolution step 1: f(x)= 7.05589\n
        differential_evolution step 2: f(x)= 5.59166\n
        differential_evolution step 3: f(x)= 2.80301\n
        ...\n
        differential_evolution step 100: f(x)= 0.538524\n

        >>> from biomass import run_simulation
        >>> optimizer.import_solution(res.x, x_id=0)
        >>> run_simulation(model, viz_type="0")
        """

        if os.path.isdir(os.path.join(self.model.path, "out", f"{x_id:d}")):
            raise ValueError(
                f"out{os.sep}{x_id:d} already exists in {self.model.path}. "
                "Use another parameter id."
            )
        else:
            os.makedirs(os.path.join(self.model.path, "out", f"{x_id:d}"))
        shutil.move(
            os.path.join(self.model.path, "out", "_external", "optimization.log"),
            os.path.join(self.model.path, "out", f"{x_id:d}"),
        )

        best_fitness: float = self.model.problem.objective(x)
        n_iter: int = 0
        with open(
            os.path.join(self.model.path, "out", f"{x_id:d}", "optimization.log"),
            mode="r",
            encoding="utf-8",
        ) as f:
            log_file = f.readlines()
        for message in log_file:
            if len(message.strip()) > 0:
                n_iter += 1
        np.save(
            os.path.join(self.model.path, "out", f"{x_id:d}", "best_fitness"),
            best_fitness,
        )
        np.save(
            os.path.join(self.model.path, "out", f"{x_id:d}", "count_num"),
            n_iter,
        )
        np.save(
            os.path.join(self.model.path, "out", f"{x_id:d}", "generation"),
            n_iter,
        )
        np.save(
            os.path.join(self.model.path, "out", f"{x_id:d}", f"fit_param{n_iter:d}"),
            x,
        )



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "TGFb",
    "Rec",
    "TGFb_pRec",
    "S2",
    "S3",
    "S4",
    "ppS2_ppS2_ppS2",
    "ppS3_ppS3_ppS3",
    "S4_S4_S4",
    "pS2",
    "pS3",
    "ppS2",
    "ppS3",
    "ppS2_ppS2_S4",
    "ppS2_ppS2_ppS3",
    "ppS2_ppS3_ppS3",
    "ppS3_ppS3_S4",
    "ppS2_ppS3_S4",
    "ppS3_S4_S4",
    "ppS2_S4_S4",
    "gene",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "MKKK",
    "MKKK_P",
    "MKK",
    "MKK_P",
    "MKK_PP",
    "MAPK",
    "MAPK_P",
    "MAPK_PP",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "k1_synthesis",
    "k1_InsIRcom",
    "k2_InsIRcom",
    "k1_p1IRcomDeg",
    "k1_p1IRcomPhos",
    "k1_p1p2IRcomdePhos",
    "k1_IRcomPhos",
    "k1_p2IRcomdePhos",
    "k1_p2IRcomDeg",
    "k1_Insp2IRcom",
    "k2_Insp2IRcom",
    "k1_p1p2IRcomDeg",
    "k1_AKTPhos",
    "k1_pAKTdePhos",
    "k1_mTORPhos",
    "k1_pmTORdePhos",
    "k1_S6KPhos",
    "k1_pS6KdePhos",
    "k1_XPhos",
    "k1_pXdePhos",
    "k1_GSK3BPhos",
    "k1_pGSK3BdePhos",
    "k1_FoxO1Phos",
    "k1_pFoxO1dePhos",
    "k1_G6PaseSynthesis",
    "k1_G6PaseDeg",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



import os
import sys
from typing import List

from setuptools import find_packages, setup


def get_version() -> str:

    version_filepath = os.path.join(os.path.dirname(__file__), "biomass", "version.py")
    with open(version_filepath) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split()[-1][1:-1]
    assert False


def get_long_description() -> str:

    readme_filepath = os.path.join(os.path.dirname(__file__), "README.md")
    with open(readme_filepath) as f:
        return f.read()


def get_install_requires() -> List[str]:

    requirements_filepath = os.path.join(os.path.dirname(__file__), "requirements.txt")
    with open(requirements_filepath) as f:
        return f.read().splitlines()


def setup_package():
    # Python version check.
    if sys.version_info[:2] < (3, 7):
        raise RuntimeError("biomass requires at least Python version 3.7")

    setup(
        name="biomass",
        version=get_version(),
        description="A Python Framework for Modeling and Analysis of Signaling Systems",
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        license="Apache 2.0",
        author="Hiroaki Imoto",
        author_email="himoto@protein.osaka-u.ac.jp",
        url="https://github.com/biomass-dev/biomass",
        download_url="https://pypi.org/project/biomass/",
        project_urls={
            "Documentation": "https://biomass-core.readthedocs.io/en/latest/",
            "Source Code": "https://github.com/biomass-dev/biomass",
        },
        packages=find_packages(exclude=["tests", "docs"]),
        install_requires=get_install_requires(),
        extras_require={
            "dev": [
                "black>=20.8b1",
                "flake8",
                "isort",
                "pre-commit",
                "pytest",
            ],
            "docs": [
                "sphinx>=1.7",
                "sphinx_rtd_theme>=0.3",
                "sphinx_autodoc_typehints>=1.10",
                "sphinxcontrib-bibtex>=2.2",
            ],
        },
        python_requires=">=3.7",
        keywords=[
            "systems",
            "biology",
            "modeling",
            "optimization",
            "sensitivity",
            "analysis",
        ],
        classifiers=[
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Software Development",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    )


if __name__ == "__main__":
    setup_package()



import numpy as np
from scipy.spatial.distance import cosine

from .observable import Observable
from .set_search_param import SearchParam


class OptimizationProblem(Observable, SearchParam):
    def __init__(self):
        super(OptimizationProblem, self).__init__()

    @property
    def bounds(self):
        """
        Lower and upper bounds on independent variables.
        """
        search_region: np.ndarray = self.get_region()
        lb = 10 ** search_region[0]
        ub = 10 ** search_region[1]
        return tuple(zip(lb, ub))

    @staticmethod
    def _compute_objval_rss(sim_data, exp_data):
        """Return Residual Sum of Squares"""
        return np.dot((sim_data - exp_data), (sim_data - exp_data))

    @staticmethod
    def _compute_objval_cos(sim_data, exp_data):
        """Return Cosine distance"""
        return cosine(sim_data, exp_data)

    @staticmethod
    def _diff_sim_and_exp(sim_matrix, exp_dict, exp_timepoint, conditions, sim_norm_max):
        sim_val = []
        exp_val = []

        for idx, condition in enumerate(conditions):
            if condition in exp_dict.keys():
                sim_val.extend(sim_matrix[list(map(int, exp_timepoint)), idx])
                exp_val.extend(exp_dict[condition])

        return np.array(sim_val) / sim_norm_max, np.array(exp_val)

    def objective(self, indiv, *args):
        """Define an objective function to be minimized"""
        if len(args) == 0:
            (x, y0) = self.update(indiv)
        elif len(args) == 1:
            raise ValueError("not enough values to unpack (expected 2, got 1)")
        elif len(args) == 2:
            (x, y0) = args
        else:
            raise ValueError("too many values to unpack (expected 2)")

        self.set_data()

        if self.simulate(x, y0) is None:
            error = np.zeros(len(self.obs_names))
            for i, obs_name in enumerate(self.obs_names):
                if self.experiments[i] is not None:
                    error[i] = self._compute_objval_rss(
                        *self._diff_sim_and_exp(
                            self.simulations[i],
                            self.experiments[i],
                            self.get_timepoint(obs_name),
                            self.conditions,
                            sim_norm_max=1
                            if not self.normalization
                            else (
                                np.max(
                                    self.simulations[
                                        self.obs_names.index(obs_name),
                                        self.normalization[obs_name]["timepoint"],
                                        [
                                            self.conditions.index(c)
                                            for c in (
                                                self.normalization[obs_name]["condition"]
                                                if self.normalization[obs_name]["condition"]
                                                else self.conditions
                                            )
                                        ],
                                    ]
                                )
                                if self.normalization[obs_name]["timepoint"] is not None
                                else np.max(
                                    self.simulations[
                                        self.obs_names.index(obs_name),
                                        :,
                                        [
                                            self.conditions.index(c)
                                            for c in (
                                                self.normalization[obs_name]["condition"]
                                                if self.normalization[obs_name]["condition"]
                                                else self.conditions
                                            )
                                        ],
                                    ]
                                )
                            ),
                        )
                    )
            """
            error = np.zeros(16)

            norm_max = np.max(self.simulations[self.obs_names.index('Phosphorylated_MEKc')])
            error[0] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_MEKc'), self.t2, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_MEKc')]['EGF']
            )
            error[1] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_MEKc'), self.t2, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_MEKc')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('Phosphorylated_ERKc')])
            error[2] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_ERKc'), self.t2, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_ERKc')]['EGF']
            )
            error[3] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_ERKc'), self.t2, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_ERKc')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('Phosphorylated_RSKw')])
            error[4] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_RSKw'), self.t2, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_RSKw')]['EGF']
            )
            error[5] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_RSKw'), self.t2, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_RSKw')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('Phosphorylated_CREBw')])
            error[6] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_CREBw'), self.t3, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_CREBw')]['EGF']
            )
            error[7] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_CREBw'), self.t3, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_CREBw')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('dusp_mRNA')])
            error[8] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('dusp_mRNA'), self.t5, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('dusp_mRNA')]['EGF']
            )
            error[9] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('dusp_mRNA'), self.t5, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('dusp_mRNA')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('cfos_mRNA')])
            error[10] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('cfos_mRNA'), self.t4, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('cfos_mRNA')]['EGF']
            )
            error[11] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('cfos_mRNA'), self.t4, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('cfos_mRNA')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('cFos_Protein')])
            error[12] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('cFos_Protein'), self.t5, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('cFos_Protein')]['EGF']
            )
            error[13] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('cFos_Protein'), self.t5, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('cFos_Protein')]['HRG']
            )

            norm_max = np.max(self.simulations[self.obs_names.index('Phosphorylated_cFos')])
            error[14] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_cFos'), self.t2, self.conditions.index('EGF')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_cFos')]['EGF']
            )
            error[15] = self._compute_objval_rss(
                self.simulations[self.obs_names.index('Phosphorylated_cFos'), self.t2, self.conditions.index('HRG')]/norm_max, 
                self.experiments[self.obs_names.index('Phosphorylated_cFos')]['HRG']
            )
            """
            return np.sum(error)  # < 1e12
        else:
            return 1e12



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "MP",
    "MC",
    "MB",
    "PC",
    "CC",
    "PCP",
    "CCP",
    "PCC",
    "PCN",
    "PCCP",
    "PCNP",
    "BC",
    "BCP",
    "BN",
    "BNP",
    "IN",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



import numpy as np

from biomass.estimation import convert_scale, initialize_search_param

from .name2idx import C, V
from .set_model import initial_values, param_values


class SearchParam(object):
    """Specify model parameters and/or initial values to optimize."""

    def __init__(self):
        # parameters
        self.idx_params = [
            C.V1,
            C.Km1,
            C.V5,
            C.Km5,
            C.V10,
            C.Km10,
            C.n10,
            C.p11,
            C.p12,
            C.p13,
            C.V14,
            C.Km14,
            C.V15,
            C.Km15,
            C.KimDUSP,
            C.KexDUSP,
            C.V20,
            C.Km20,
            C.V21,
            C.Km21,
            C.V24,
            C.Km24,
            C.V25,
            C.Km25,
            C.KimRSK,
            C.KexRSK,
            C.V27,
            C.Km27,
            C.V28,
            C.Km28,
            C.V29,
            C.Km29,
            C.V30,
            C.Km30,
            C.V31,
            C.Km31,
            C.n31,
            C.p32,
            C.p33,
            C.p34,
            C.V35,
            C.Km35,
            C.V36,
            C.Km36,
            C.V37,
            C.Km37,
            C.KimFOS,
            C.KexFOS,
            C.V42,
            C.Km42,
            C.V43,
            C.Km43,
            C.V44,
            C.Km44,
            C.p47,
            C.m47,
            C.p48,
            C.p49,
            C.m49,
            C.p50,
            C.p51,
            C.m51,
            C.V57,
            C.Km57,
            C.n57,
            C.p58,
            C.p59,
            C.p60,
            C.p61,
            C.KimF,
            C.KexF,
            C.p63,
            C.KF31,
            C.nF31,
            C.a,
        ]

        # initial values
        self.idx_initials = [
            # V.(specie)
        ]

    def get_region(self):
        x = param_values()
        y0 = initial_values()

        search_param = initialize_search_param(
            parameters=C.NAMES,
            species=V.NAMES,
            param_values=x,
            initial_values=y0,
            estimated_params=self.idx_params,
            estimated_initials=self.idx_initials,
        )

        search_rgn = np.zeros((2, len(x) + len(y0)))
        # Default: 0.1 ~ 10
        for i, j in enumerate(self.idx_params):
            search_rgn[0, j] = search_param[i] * 0.1  # lower bound
            search_rgn[1, j] = search_param[i] * 10.0  # upper bound
        # Default: 0.5 ~ 2
        for i, j in enumerate(self.idx_initials):
            search_rgn[0, j + len(x)] = search_param[i + len(self.idx_params)] * 0.5  # lower bound
            search_rgn[1, j + len(x)] = search_param[i + len(self.idx_params)] * 2.0  # upper bound

        # search_rgn[:,C.parameter] = [lower_bound, upper_bound]
        # search_rgn[:,V.specie+len(x)] = [lower_bound, upper_bound]

        search_rgn[:, C.V1] = [7.33e-2, 6.60e-01]
        search_rgn[:, C.Km1] = [1.83e2, 8.50e2]
        search_rgn[:, C.V5] = [6.48e-3, 7.20e1]
        search_rgn[:, C.Km5] = [6.00e-1, 1.60e04]
        search_rgn[:, C.V10] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km10] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.n10] = [1.00, 4.00]
        search_rgn[:, C.p11] = [8.30e-13, 1.44e-2]
        search_rgn[:, C.p12] = [8.00e-8, 5.17e-2]
        search_rgn[:, C.p13] = [1.38e-7, 4.84e-1]
        search_rgn[:, C.V14] = [4.77e-3, 4.77e1]
        search_rgn[:, C.Km14] = [2.00e2, 2.00e6]
        search_rgn[:, C.V15] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km15] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.KimDUSP] = [2.20e-4, 5.50e-1]
        search_rgn[:, C.KexDUSP] = [2.60e-4, 6.50e-1]
        search_rgn[:, C.V20] = [4.77e-3, 4.77e1]
        search_rgn[:, C.Km20] = [2.00e2, 2.00e6]
        search_rgn[:, C.V21] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km21] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.V24] = [4.77e-2, 4.77e0]
        search_rgn[:, C.Km24] = [2.00e3, 2.00e5]
        search_rgn[:, C.V25] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km25] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.KimRSK] = [2.20e-4, 5.50e-1]
        search_rgn[:, C.KexRSK] = [2.60e-4, 6.50e-1]
        search_rgn[:, C.V27] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km27] = [1.00e2, 1.00e4]
        search_rgn[:, C.V28] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km28] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.V29] = [4.77e-2, 4.77e0]
        search_rgn[:, C.Km29] = [2.93e3, 2.93e5]
        search_rgn[:, C.V30] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km30] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.V31] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km31] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.n31] = [1.00, 4.00]
        search_rgn[:, C.p32] = [8.30e-13, 1.44e-2]
        search_rgn[:, C.p33] = [8.00e-8, 5.17e-2]
        search_rgn[:, C.p34] = [1.38e-7, 4.84e-1]
        search_rgn[:, C.V35] = [4.77e-3, 4.77e1]
        search_rgn[:, C.Km35] = [2.00e2, 2.00e6]
        search_rgn[:, C.V36] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km36] = [1.00e2, 1.00e4]
        search_rgn[:, C.V37] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km37] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.KimFOS] = [2.20e-4, 5.50e-1]
        search_rgn[:, C.KexFOS] = [2.60e-4, 6.50e-1]
        search_rgn[:, C.V42] = [4.77e-3, 4.77e1]
        search_rgn[:, C.Km42] = [2.00e2, 2.00e6]
        search_rgn[:, C.V43] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km43] = [1.00e2, 1.00e4]
        search_rgn[:, C.V44] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km44] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.p47] = [1.45e-4, 1.45e0]
        search_rgn[:, C.m47] = [6.00e-3, 6.00e1]
        search_rgn[:, C.p48] = [2.70e-3, 2.70e1]
        search_rgn[:, C.p49] = [5.00e-5, 5.00e-1]
        search_rgn[:, C.m49] = [5.00e-3, 5.00e1]
        search_rgn[:, C.p50] = [3.00e-3, 3.00e1]
        search_rgn[:, C.p51] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.m51] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.V57] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.Km57] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.n57] = [1.00, 4.00]
        search_rgn[:, C.p58] = [8.30e-13, 1.44e-2]
        search_rgn[:, C.p59] = [8.00e-8, 5.17e-2]
        search_rgn[:, C.p60] = [1.38e-7, 4.84e-1]
        search_rgn[:, C.p61] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.KimF] = [2.20e-4, 5.50e-1]
        search_rgn[:, C.KexF] = [2.60e-4, 6.50e-1]
        search_rgn[:, C.p63] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.KF31] = [np.exp(-10), np.exp(10)]
        search_rgn[:, C.nF31] = [1.00, 4.00]
        search_rgn[:, C.a] = [1.00e2, 5.00e2]

        search_rgn = convert_scale(
            region=search_rgn,
            parameters=C.NAMES,
            species=V.NAMES,
            estimated_params=self.idx_params,
            estimated_initials=self.idx_initials,
        )

        return search_rgn

    def update(self, indiv):
        x = param_values()
        y0 = initial_values()

        for i, j in enumerate(self.idx_params):
            x[j] = indiv[i]
        for i, j in enumerate(self.idx_initials):
            y0[j] = indiv[i + len(self.idx_params)]

        # constraints ----------------------------------------------------------
        x[C.V6] = x[C.V5]
        x[C.Km6] = x[C.Km5]
        x[C.KimpDUSP] = x[C.KimDUSP]
        x[C.KexpDUSP] = x[C.KexDUSP]
        x[C.KimpcFOS] = x[C.KimFOS]
        x[C.KexpcFOS] = x[C.KexFOS]
        x[C.p52] = x[C.p47]
        x[C.m52] = x[C.m47]
        x[C.p53] = x[C.p48]
        x[C.p54] = x[C.p49]
        x[C.m54] = x[C.m49]
        x[C.p55] = x[C.p50]
        x[C.p56] = x[C.p51]
        x[C.m56] = x[C.m51]
        # ----------------------------------------------------------------------

        return x, y0

    def gene2val(self, indiv_gene):
        search_rgn = self.get_region()
        indiv = 10 ** (indiv_gene * (search_rgn[1, :] - search_rgn[0, :]) + search_rgn[0, :])

        return indiv

    def val2gene(self, indiv):
        search_rgn = self.get_region()
        indiv_gene = (np.log10(indiv) - search_rgn[0, :]) / (search_rgn[1, :] - search_rgn[0, :])

        return indiv_gene



import os
from dataclasses import dataclass
from math import isnan
from typing import List, Optional

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes._axes import _log as matplotlib_axes_logger

from ..exec_model import ExecModel, ModelObject
from ..plotting import MultipleObservables, SingleObservable

matplotlib_axes_logger.setLevel("ERROR")


@dataclass
class TemporalDynamics(ExecModel):
    model: ModelObject

    def plot_timecourse(
        self,
        n_file: List[int],
        viz_type: str,
        show_all: bool,
        stdev: bool,
        simulations_all: np.ndarray,
    ) -> None:
        """
        Plot time course of each observable.

        Parameters
        ----------
        n_file : list of integers
            Optimized parameter sets in out/.

        viz_type : str
            One of ['average', 'best', 'original', 'n(=1,2,...)', 'experiment'].

        show_all : bool
            Whether to show all simulation results.

        stdev : bool
            If True, the standard deviation of simulated values will be shown
            (only available for 'average' visualization type).

        simulations_all : numpy array
            Array containing all simulated values.

        """
        os.makedirs(
            os.path.join(
                self.model.path,
                "figure",
                "simulation",
                f"{viz_type}",
            ),
            exist_ok=True,
        )
        self.model.problem.set_data()
        self.model.viz.set_timecourse_rcParams()
        singleplotting = self.model.viz.get_single_observable_options()
        multiplotting = self.model.viz.get_multiple_observables_options()

        for mode in range(2):
            # mode 0 : timecourse_for_each_observable
            # mode 1 : multiple_observables
            set_fig = False
            for i, obs_name in enumerate(self.model.observables):
                if len(singleplotting[i].cmap) < len(self.model.problem.conditions) or len(
                    singleplotting[i].shape
                ) < len(self.model.problem.conditions):
                    raise ValueError(
                        "len(cmap), len(shape) must be equal to"
                        " or greater than len(self.model.problem.conditions)."
                    )
                if mode == 1 and obs_name not in multiplotting.observables:
                    continue
                if mode == 0:
                    plt.figure(figsize=singleplotting[i].figsize)
                elif mode == 1 and not set_fig:
                    plt.figure(figsize=multiplotting.figsize)
                    set_fig = True
                plt.gca().spines["right"].set_visible(False)
                plt.gca().spines["top"].set_visible(False)
                if viz_type != "experiment":
                    if show_all:
                        self._plot_show_all(
                            n_file, simulations_all, obs_name, mode, singleplotting, multiplotting
                        )
                    if viz_type == "average":
                        normalized = self._normalize_array(
                            n_file, simulations_all, obs_name, mode, singleplotting, multiplotting
                        )
                        if (
                            self.model.problem.normalization
                            and self.model.problem.normalization[obs_name]["timepoint"] is None
                        ):
                            normalized = self._divide_by_maximum(normalized, obs_name)
                        self._plot_average(normalized, obs_name, mode, singleplotting, multiplotting)
                        if stdev:
                            self._show_sd(normalized, obs_name, mode, singleplotting, multiplotting)
                    else:
                        self._plot_simulations(obs_name, mode, singleplotting, multiplotting)
                if (
                    viz_type == "experiment" or singleplotting[i].exp_data
                ) and self.model.problem.experiments[i] is not None:
                    exp_t = self.model.problem.get_timepoint(obs_name)
                    if self.model.problem.error_bars[i] is not None:
                        self._plot_experimental_data_with_error_bars(
                            viz_type, exp_t, obs_name, mode, singleplotting, multiplotting
                        )
                    else:
                        self._plot_experimental_data_without_error_bars(
                            viz_type, exp_t, obs_name, mode, singleplotting, multiplotting
                        )
                if mode == 0:
                    self._save_mode_0(obs_name, singleplotting, viz_type)
            if mode == 1 and multiplotting.observables:
                self._save_mode_1(multiplotting, viz_type)

    def _plot_show_all(
        self,
        n_file: List[int],
        simulations_all: np.ndarray,
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> None:
        """
        Plot time course simulated values (show_all == True).
        """
        i = self.model.observables.index(obs_name)
        for j, _ in enumerate(n_file):
            for l, condition in enumerate(self.model.problem.conditions):
                if (mode == 0 and condition not in singleplotting[i].dont_show) or (
                    mode == 1 and condition == multiplotting.condition
                ):
                    plt.plot(
                        np.array(self.model.problem.t) / singleplotting[i].divided_by,
                        simulations_all[i, j, :, l]
                        / (
                            1
                            if not self.model.problem.normalization
                            or np.max(simulations_all[i, j, :, l]) == 0.0
                            else np.max(
                                simulations_all[
                                    i,
                                    j,
                                    self.model.problem.normalization[obs_name]["timepoint"],
                                    [
                                        self.model.problem.conditions.index(c)
                                        for c in self.model.problem.normalization[obs_name][
                                            "condition"
                                        ]
                                    ],
                                ]
                            )
                            if self.model.problem.normalization[obs_name]["timepoint"] is not None
                            else np.max(
                                simulations_all[
                                    i,
                                    j,
                                    :,
                                    [
                                        self.model.problem.conditions.index(c)
                                        for c in self.model.problem.normalization[obs_name][
                                            "condition"
                                        ]
                                    ],
                                ]
                            )
                        ),
                        color=singleplotting[i].cmap[l]
                        if mode == 0
                        else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                        alpha=0.05,
                    )

    def _normalize_array(
        self,
        n_file: List[int],
        simulations_all: np.ndarray,
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> np.ndarray:
        """
        Normalize the array simulations_all using problem.normalization set in observable.py.
        """
        normalized: np.ndarray = np.empty_like(simulations_all)
        i = self.model.observables.index(obs_name)
        for j, _ in enumerate(n_file):
            for l, condition in enumerate(self.model.problem.conditions):
                if (mode == 0 and condition not in singleplotting[i].dont_show) or (
                    mode == 1 and condition == multiplotting.condition
                ):
                    normalized[i, j, :, l] = simulations_all[i, j, :, l] / (
                        1
                        if not self.model.problem.normalization
                        or np.max(simulations_all[i, j, :, l]) == 0.0
                        else np.max(
                            simulations_all[
                                i,
                                j,
                                self.model.problem.normalization[obs_name]["timepoint"],
                                [
                                    self.model.problem.conditions.index(c)
                                    for c in self.model.problem.normalization[obs_name][
                                        "condition"
                                    ]
                                ],
                            ]
                        )
                        if self.model.problem.normalization[obs_name]["timepoint"] is not None
                        else np.max(
                            simulations_all[
                                i,
                                j,
                                :,
                                [
                                    self.model.problem.conditions.index(c)
                                    for c in self.model.problem.normalization[obs_name][
                                        "condition"
                                    ]
                                ],
                            ]
                        )
                    )
        return normalized

    def _divide_by_maximum(self, normalized: np.ndarray, obs_name: str) -> np.ndarray:
        """
        Divide the array by its maximum.
        """
        mean_vec = []
        for c in self.model.problem.normalization[obs_name]["condition"]:
            mean_vec.append(
                np.nanmean(
                    normalized[
                        self.model.observables.index(obs_name),
                        :,
                        :,
                        self.model.problem.conditions.index(c),
                    ],
                    axis=0,
                )
            )
        norm_max = np.max(mean_vec)
        if not isnan(norm_max) and norm_max != 0.0:
            normalized[self.model.observables.index(obs_name), :, :, :] /= norm_max

        return normalized

    def _plot_average(
        self,
        normalized: np.ndarray,
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> None:
        """
        Plot time course simulated values (viz_type == 'average').
        """
        i = self.model.observables.index(obs_name)
        for l, condition in enumerate(self.model.problem.conditions):
            if (mode == 0 and condition not in singleplotting[i].dont_show) or (
                mode == 1 and condition == multiplotting.condition
            ):
                plt.plot(
                    np.array(self.model.problem.t) / singleplotting[i].divided_by,
                    np.nanmean(normalized[i, :, :, l], axis=0),
                    color=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    label=condition if mode == 0 else singleplotting[i].ylabel,
                )

    def _show_sd(
        self,
        normalized: np.ndarray,
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> None:
        """
        Plot standard deviation (SD) as shaded area when stdev == True.
        """
        i = self.model.observables.index(obs_name)
        for l, condition in enumerate(self.model.problem.conditions):
            if (mode == 0 and condition not in singleplotting[i].dont_show) or (
                mode == 1 and condition == multiplotting.condition
            ):
                y_mean = np.nanmean(normalized[i, :, :, l], axis=0)
                y_std = [
                    np.nanstd(normalized[i, :, k, l], ddof=1)
                    for k, _ in enumerate(self.model.problem.t)
                ]
                plt.fill_between(
                    np.array(self.model.problem.t) / singleplotting[i].divided_by,
                    y_mean - y_std,
                    y_mean + y_std,
                    lw=0,
                    color=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    alpha=0.1,
                )

    def _plot_simulations(
        self,
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> None:
        """
        Plot time course simulated values (viz_type not in ['average', 'experiment']).
        """
        i = self.model.observables.index(obs_name)
        for l, condition in enumerate(self.model.problem.conditions):
            if (mode == 0 and condition not in singleplotting[i].dont_show) or (
                mode == 1 and condition == multiplotting.condition
            ):
                plt.plot(
                    np.array(self.model.problem.t) / singleplotting[i].divided_by,
                    self.model.problem.simulations[i, :, l]
                    / (
                        1
                        if not self.model.problem.normalization
                        or np.max(self.model.problem.simulations[i, :, l]) == 0.0
                        else np.max(
                            self.model.problem.simulations[
                                i,
                                self.model.problem.normalization[obs_name]["timepoint"],
                                [
                                    self.model.problem.conditions.index(c)
                                    for c in self.model.problem.normalization[obs_name][
                                        "condition"
                                    ]
                                ],
                            ]
                        )
                        if self.model.problem.normalization[obs_name]["timepoint"] is not None
                        else np.max(
                            self.model.problem.simulations[
                                i,
                                :,
                                [
                                    self.model.problem.conditions.index(c)
                                    for c in self.model.problem.normalization[obs_name][
                                        "condition"
                                    ]
                                ],
                            ]
                        )
                    ),
                    color=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    label=condition if mode == 0 else singleplotting[i].ylabel,
                )

    def _plot_experimental_data_with_error_bars(
        self,
        viz_type: str,
        exp_t: Optional[List[int]],
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> None:
        """
        Plot experimental measurements with error bars.
        """
        i = self.model.observables.index(obs_name)
        for l, condition in enumerate(self.model.problem.conditions):
            if (
                condition in self.model.problem.experiments[i]
                and (mode == 0 and condition not in singleplotting[i].dont_show)
                or (mode == 1 and condition == multiplotting.condition)
            ):
                exp_data = plt.errorbar(
                    np.array(exp_t) / singleplotting[i].divided_by,
                    self.model.problem.experiments[i][condition],
                    yerr=self.model.problem.error_bars[i][condition],
                    color=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    ecolor=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    elinewidth=1,
                    capsize=8,
                    markerfacecolor="None",
                    markeredgecolor=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    fmt=singleplotting[i].shape[l]
                    if mode == 0
                    else multiplotting.shape[multiplotting.observables.index(obs_name)],
                    clip_on=False,
                    label=singleplotting[i].ylabel
                    if mode == 1 and viz_type == "experiment"
                    else None,
                )
                for capline in exp_data[1]:
                    capline.set_clip_on(False)
                for barlinecol in exp_data[2]:
                    barlinecol.set_clip_on(False)

    def _plot_experimental_data_without_error_bars(
        self,
        viz_type: str,
        exp_t: Optional[List[int]],
        obs_name: str,
        mode: int,
        singleplotting: List[SingleObservable],
        multiplotting: MultipleObservables,
    ) -> None:
        """
        Plot experimental measurements when model.problem.error_bars[i] is None.
        """
        i = self.model.observables.index(obs_name)
        for l, condition in enumerate(self.model.problem.conditions):
            if (
                condition in self.model.problem.experiments[i]
                and (mode == 0 and condition not in singleplotting[i].dont_show)
                or (mode == 1 and condition == multiplotting.condition)
            ):
                plt.plot(
                    np.array(exp_t) / singleplotting[i].divided_by,
                    self.model.problem.experiments[i][condition],
                    singleplotting[i].shape[l]
                    if mode == 0
                    else multiplotting.shape[multiplotting.observables.index(obs_name)],
                    markerfacecolor="None",
                    markeredgecolor=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    color=singleplotting[i].cmap[l]
                    if mode == 0
                    else multiplotting.cmap[multiplotting.observables.index(obs_name)],
                    clip_on=False,
                    label=singleplotting[i].ylabel
                    if mode == 1 and viz_type == "experiment"
                    else None,
                )

    def _save_mode_0(
        self,
        obs_name: str,
        singleplotting: List[SingleObservable],
        viz_type: str,
    ) -> None:
        """
        Plot time course of each observable.
        """
        i = self.model.observables.index(obs_name)
        if singleplotting[i].xlim:
            plt.xlim(singleplotting[i].xlim)
        if singleplotting[i].xticks is not None:
            plt.xticks(singleplotting[i].xticks)
        plt.xlabel(singleplotting[i].xlabel)
        if singleplotting[i].ylim:
            plt.ylim(singleplotting[i].ylim)
        if singleplotting[i].yticks is not None:
            plt.yticks(singleplotting[i].yticks)
        plt.ylabel(singleplotting[i].ylabel)
        if singleplotting[i].legend_kws is not None:
            plt.legend(**singleplotting[i].legend_kws)
        plt.savefig(
            os.path.join(
                self.model.path,
                "figure",
                "simulation",
                f"{viz_type}",
                f"{obs_name}",
            ),
        )
        plt.close()

    def _save_mode_1(
        self,
        multiplotting: MultipleObservables,
        viz_type: str,
    ) -> None:
        """
        Plot time course of multiple observables in one figure.
        """
        if multiplotting.xlim:
            plt.xlim(multiplotting.xlim)
        if multiplotting.xticks is not None:
            plt.xticks(multiplotting.xticks)
        plt.xlabel(multiplotting.xlabel)
        if multiplotting.ylim:
            plt.ylim(multiplotting.ylim)
        if multiplotting.yticks is not None:
            plt.yticks(multiplotting.yticks)
        plt.ylabel(multiplotting.ylabel)
        plt.legend(**multiplotting.legend_kws)
        plt.savefig(
            os.path.join(
                self.model.path,
                "figure",
                "simulation",
                f"{viz_type}",
                f"{multiplotting.fname}",
            ),
        )
        plt.close()
